#include <stdio.h>
#include <setjmp.h>
#include <memory.h>
#include <string.h>
#ifdef M_XENIX
# include <signal.h>
#else
# include <sys/signal.h>
#endif
#include "config.h"
#include "lint.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "lang.h"
#include "realms.h"
 
/* this file is for error recovery using the setjmp() 
 * it is important to maintain certan global variables, after an error
 * as they were before the error.

 * A few, we just zero out.  These are used when compiling, and have no
 * regular value any other time.
 */


void smart_log();
void abort();
int kill(),getpid();
void add_message();
void dump_trace();

struct jmp_globals {
	struct lnode_2 *next_arg_list_to_use;
	char case_match,break_flag,break_level;
#ifdef AIX
	signed char trace_depth;
#else
	char trace_depth;
#endif
	int current_line;
	LNODE *case_def;
	VALUE *switch_arg;
	OBJECT *command_giver;
	struct object_status *ob_status[2];
};

extern char new_error;
extern struct lnode_2 *next_arg_list_to_use;
extern char case_match,break_flag,break_level;
#ifdef AIX
extern signed char trace_depth;
#else
extern char trace_depth;
#endif
extern int current_line;
extern LNODE *case_def;
extern VALUE *switch_arg;
extern struct object_status *ob_status[2];
extern OBJECT *command_giver;
extern char *compile_file;

jmp_buf error_recovery_context;

void set_jump_state(erc,jmp)
   struct jmp_globals *jmp;
   jmp_buf erc;
{
  if(jmp)
   {
	jmp->trace_depth=trace_depth;
	jmp->ob_status[0]=ob_status[0];
	jmp->ob_status[1]=ob_status[1];
	jmp->command_giver=command_giver;
	jmp->case_match=case_match;
	jmp->case_def=case_def;
	jmp->next_arg_list_to_use=next_arg_list_to_use;
	jmp->break_flag=break_flag;
	jmp->break_level=break_level;
	jmp->current_line=current_line;
   }

  if(erc)
   	memcpy((char *) erc,(char *) error_recovery_context,
		   sizeof error_recovery_context);
}

void clear_jump_state(jmp)
	struct jmp_globals *jmp;
{
	if(ob_status[0] && ob_status[0]!=jmp->ob_status[0] &&
	   ob_status[0]!=jmp->ob_status[1]) free_ob_status(ob_status[0]);
	if(ob_status[1] && ob_status[1]!=jmp->ob_status[0] &&
	   ob_status[1]!=ob_status[1]) free_ob_status(ob_status[1]);
	ob_status[0]=jmp->ob_status[0];
	ob_status[1]=jmp->ob_status[1];
	command_giver=jmp->command_giver;
	case_match=jmp->case_match;
	case_def=jmp->case_def;
	next_arg_list_to_use=jmp->next_arg_list_to_use;
	break_flag=jmp->break_flag;
	break_level=jmp->break_level;
	current_line=jmp->current_line;
	trace_depth=jmp->trace_depth;
        new_error=1;
}

void reset_jump_state(erc)
	jmp_buf erc;
{
    memcpy((char *) error_recovery_context,(char *) erc,
		   sizeof error_recovery_context);
}
 
char in_fatal=0;

/*VARARGS1*/
void fatal(fmt, a, b, c, d, e, f, g, h)
    char *fmt;
    int a, b, c, d, e, f, g, h;
{
    /* Prevent double fatal. */
          (void)signal(SIGHUP , SIG_DFL);
          (void)signal(SIGTERM, SIG_DFL);
          (void)signal(SIGKILL, SIG_DFL);
          (void)signal(SIGILL , SIG_DFL);
          (void)signal(SIGIOT , SIG_DFL);
#ifndef AIX
          (void)signal(SIGABRT, SIG_DFL);
#endif

#ifndef sun
          (void)signal(SIGSEGV, SIG_DFL);
#endif
#if 0
    if (in_fatal) (void)abort();
#else
   if(in_fatal) kill(getpid(),SIGIOT);
#endif

    debug_message(sys_realm,fmt, a, b, c, d, e, f, g, h);
    if (ob_status[0]->ob)
		debug_message(sys_realm,"Current object was %s\n", ob_status[0]->ob->name);
    dump_trace(sys_realm);
   in_fatal=1;
   kill(getpid(),SIGIOT); 
}

static char emsg_buf[2000];
int num_error = 0;

/*VARARGS1*/
void error(realm,fmt, a, b, c, d, e, f, g, h)
    struct realms *realm;
    char *fmt;
    int a, b, c, d, e, f, g, h;
{
    extern struct object *current_heart_beat;
    extern struct value * catch_value;
    char s_buff[256];
    char pat[256];
    char *object_name;

    extern char *current_file;


    if(!current_file && ob_status[0]->ob)
     {
       (void)sprintf(pat,"%s %s","%s:(%d) : ",fmt);
       (void)sprintf(emsg_buf,pat,ob_status[0]->ob->name,current_line,a,b,c,d,e,f,g,h);
     }
    else
      {
        if(current_file) free(current_file);
	current_file=0;
        (void)sprintf(emsg_buf, fmt, a, b, c, d, e, f, g, h);
      }

    num_error++;
    if (num_error > 1) fatal("Too many simultaneous errors.\n");

    debug_message(realm,emsg_buf);


    if (ob_status[0]->ob) 
      {
		extern void save_error PROT((char *,char *,int));

		if(command_giver)
    			 add_message("Error while executing program in object '%s'.\n",ob_status[0]->ob->name);

	    smart_log(ob_status[0]->ob->name,current_line,emsg_buf,1);
	    save_error(emsg_buf, ob_status[0]->ob->name, current_line);
      }
    dump_trace(realm);
    trace_depth = -1;
    num_error--;
    longjmp(error_recovery_context, 1);
    (void)abort();
}

void smart_log(error_file, line, what,flag)
     char *error_file, *what;
     int line;
     int flag;
{
	extern REALMS *extract_realm PROT((char *));
	extern char *extract_wizard PROT((char *));
	extern void log_file PROT((char *,char *,char *));

    char buff[200], *p;

    if (!error_file || error_file[0]==0) return;
	p=extract_wizard(flag?error_file:compile_file);
    if (!p) return;
    if(!flag)
     {
    	if(strcmp(error_file,compile_file))
	 {
      	  (void)sprintf(buff,"In file: %s\n",compile_file);
	  log_file(p,buff,(extract_realm(compile_file))->realm_dir);
	 }
    	(void)sprintf(buff, "%s:(%d) : %s\n", error_file, line, what);
     }
    else (void)sprintf(buff,"%s",what);
    log_file(p, buff,(extract_realm(flag?error_file:compile_file))->realm_dir);

    free(p);
}

/*
 * Write out a trace. If there is an heart_beat(), then return the
 * object that had that heart beat.
 */
void dump_trace(realm) 
   struct realms *realm;
{
    extern struct trace trace_back[];

    int i;
    char s_buff[200];

    if(trace_depth!=-1)
 	{
	
           debug_message(realm,"-------------------------------- Dump Trace ---------------------------------\n");
	   for (i=0; i<=trace_depth; i++) 
	   {
	   int line;

	   line=trace_back[i].line;

	   (void)sprintf(s_buff,"%2d:'%-21s' in '%-35s' line %d\n", i,
                 trace_back[i].function_name, trace_back[i].prog_name,line);
           debug_message(realm,"%s",s_buff);

  	   if(ob_status[0]->ob)
		 smart_log(ob_status[0]->ob->name,line,s_buff,1); 
	   }
	}
}

void do_error(num_arg,fun)
  char *num_arg,*fun;
{
	  if(fun[0]!='\'') error(ob_status[0]->ob->realm,"Bad %s argument to %s()\n",num_arg,fun);
	     else error(ob_status[0]->ob->realm,"Bad %s argument to %s\n",num_arg,fun);
}
