union u {
	int n_int;
        double n_float;
	char *string;
	struct object *ob;
	struct vector *vec;
};
/*
 * The definition of a value node. It is used for all computations.
 * Strings can be constant, in which case they must not be freed.
 */
struct value {
    unsigned char type;
    unsigned char constant;		/* This string is not alloced with malloc */
    struct value *next;		/* Linked list of values. */
    union u u;
};

/*
 * This structure is equivalent with struct value but with one exception.
 * The 'next' pointer and the constant flag is not here. This will save
 * space for all global variables in objects, as the pointer is not needed
 * for them.
 */
struct svalue {
    unsigned char type;
    union u u;
};

#define T_INVALID	0x000
#define T_STRING	0x001
#define T_POINTER	0x002
#define T_OBJECT	0x004
#define T_FLOAT 	0x011
#define T_INT		0x012

#define T_NUMBER	0x010

struct value *alloc_value();
struct svalue *find_value PROT((struct lnode_number *));
struct value	*make_number PROT((int,double,int)), 
		*make_object PROT((struct object *)),
		*make_string PROT ((char *));

extern struct value const0, const1;

struct vector {
    unsigned short size;
    unsigned short real_size;
    short ref;
    struct wiz_list *wl;	/* Save who made the vector */
    unsigned char sorted;
    struct svalue item[1];
};


struct object_status {
	struct object *ob;
	struct inherit_list *ip;
        struct value **local_names;
	struct value *arg;
};

struct trace {
    char *function_name;
    char *object_name;
    char *prog_name;
    char *command_name;
    int line;
};

struct fun_p {
	struct lnode_def *fun_p;
	struct inherit_list *ip;
};

#define ALLOC_VECTOR(nelem) \
(struct vector *)xalloc((int)(sizeof (struct vector) + sizeof(struct svalue) * (nelem - 1)))

void free_vector PROT((struct vector *)), free_all_values(),
     call_function PROT((struct lnode_def *));

void free_ob_status PROT((struct object_status *));

struct object_status *alloc_ob_status PROT((void));
