%{
/*
 * This is the grammar definition of ROC. Expect 1 shift/reduce conflict,
 * because of if-then-else. The default, to shift this conflict, is the
 * correct solution.
 */
#include <string.h>
#include <stdio.h>
#include "config.h"
#include "lint.h"
#include "lnode.h"
#include "object.h"
#include "error.h"
#include "realms.h"
#include "lang.h"
#include "lan_fun.h"
#include "lex.h"
#include "s_hash.h"
#include "p_hash.h"
#include "wiz_list.h"
#include "var_hash.h"
#include "lint2.h"

#define YYMAXDEPTH	600

#ifdef __GNUC__
#define free(p)  (free((char *)p))
#endif


#ifndef hp300
char *strstr();
#endif

struct lnode *prog;
int syntax_error=0;
unsigned char function_flags,fun_only;
unsigned short variable_count;
void free_all_local_names(), add_local_name(),yyerror();
extern int yylex();
static void copy_variables();
static int check_deklared(),check_declared();
extern int current_line;
unsigned short object_flags;

char *argument_name;	/* The name of the current argument. */
char *xalloc(), *string_copy();
unsigned char var_flags,no_block;
extern struct object_status *ob_status[];
extern char *compile_file;
char *var_fun_name;

/*
 * 'inherit_file' is used as a flag. If it is set to a string
 * after yyparse(), this string should be loaded as an object,
 * and the original object must be loaded again.
 */
extern char *current_file, *inherit_file;

extern struct inherit_list *new_inherit_list;

char **local_names;

unsigned char current_number_of_locals = 0;
unsigned char num_arg;
struct lnode_def *current_functions = 0;
%}

%token F_CASE F_CONS F_DEFAULT F_GLOBAL F_IF F_FUNC_DECL F_IDENTIFIER F_LAND F_LOR
%token F_OBJECT F_SET_TRUE F_SET_FALSE F_RETURN F_NOT F_POINTER F_PRIVATE
%token F_STRING F_SWITCH F_INC F_DEC F_POST_INC F_POST_DEC F_COMMA F_HASH
%token F_ASSIGN F_INT F_ADD F_SUBTRACT F_MULTIPLY F_DIVIDE F_LT F_GT
%token F_EQ F_GE F_LE F_ARGUMENT F_FUNCTION F_NE F_ADD_EQ F_SUB_EQ F_DIV_EQ
%token F_MULT_EQ F_NEGATE F_SUBSCRIPT F_WHILE F_BREAK F_DO F_FOR F_STRING_DECL
%token F_LOCAL_NAME F_ELSE F_CONST0 F_CONST1 F_BLOCK F_CONTINUE F_INT_DECL
%token F_MOD F_MOD_EQ F_INHERIT F_COLON_COLON F_STATIC F_ARROW F_AGGREGATE
%token F_COMPL F_AND F_AND_EQ F_OR F_OR_EQ F_XOR F_XOR_EQ F_LSH F_LSH_EQ F_RSH
%token F_RSH_EQ F_NO_EXT_SAVE F_RESET F_NO_CLONE F_FLOAT F_TEST_PLAYER
%token F_ADD_ACTION F_ADD_WORTH F_ALL_INVENTORY F_ALLOCATE F_BLOCK_SHOUTS
%token F_CALL_OTHER F_CALL_OUT F_CAPITALIZE F_CAT F_CLEAR_BIT F_FLOAT_DECL
%token F_CLONE_OBJECT F_COMBINE_FREE_LIST F_COMMAND F_CP F_GREP F_DOT_DOT
%token F_CREATOR F_CRYPT F_CTIME F_DEL_ELEMENT F_DESTRUCT F_SEARCH_ARRAY
%token F_ED F_ENABLE_COMMANDS F_ENVIRONMENT F_EXPLODE F_EXTRACT F_FILE_NAME
%token F_FILE_SIZE F_FIND_LINK F_FIND_LIVING F_FIND_OBJECT F_MIXED F_VOID
%token F_NO_MASK F_FIND_PLAYER F_OBJECT_NAME F_ITOC F_STRSTR F_NTOA
%token F_FIRST_INVENTORY F_IMPLODE F_INPUT_TO F_NUMBERP F_LIVING F_LOCALCMD
%token F_LOG_FILE F_LOG_OBJECT F_LOWER_CASE F_LS F_MKDIR F_MOVE_OBJECT F_MV
%token F_NEW_PLAYER F_NEXT_INVENTORY F_NOTIFY_FAIL F_OBJECTP F_QUERY_PORT_NUMBER
%token F_POINTERP F_PRESENT F_PREVIOUS_OBJECT F_QUERY_CALL_OUT F_QUERY_ED
%token F_QUERY_HOST_NAME F_ABS F_ATOF F_ATOI F_CEIL F_EXP F_FABS F_FLOOR
%token F_FMOD F_QUERY_IDLE F_QUERY_IP_NUMBER F_QUERY_LIGHT F_QUERY_LOAD_AVERAGE
%token F_QUERY_PROMPT F_QUERY_REALM F_QUERY_SNOOP F_QUERY_VERB F_RANDOM
%token F_READ_FILE F_REALM_LIST F_REMOVE_ALL_VERBS F_REMOVE_CALL_OUT
%token F_REMOVE_VERB F_RESTORE_INVENTORY F_POW F_LAST_INVENTORY F_SQRT
%token F_RESTORE_OBJECT F_RM F_RMDIR F_SAVE_INVENTORY F_BRACED_EXTRACT
%token F_SAVE_OBJECT F_SAY F_SET_BIT F_SET_HEART_BEAT F_SET_LIGHT F_SYSTEM
%token F_SET_LIVING_NAME F_SET_PROMPT F_SET_REALM F_SHOUT F_PREVIOUS_INVENTORY
%token F_SHUTDOWN F_SHUTTING_DOWN F_SIZEOF F_SLICE_ARRAY F_SNOOP 
%token F_SORT_ARRAY F_SPRINTF F_BSEARCH_ARRAY F_ADD_SORT
%token F_SSCANF F_STRCHR F_STRINGP F_STRLEN F_STRNCMP
%token F_STRRCHR F_SWAP F_TAIL F_TELL_OBJECT F_TELL_ROOM F_TEST_BIT
%token F_THIS_OBJECT F_CAST F_TYPE F_ADD_REALM
%token F_THIS_PLAYER F_TIME F_UPPER_CASE F_USED_OBJECT
%token F_USERS F_VERB_MATCHING F_VERSION F_WIZLIST F_WRITE F_WRITE_FILE

%union
{
	struct lnode *lnode;
	int number;
	char *string;
}

%type <lnode> cond expr1 expr_list expr2 expr3 function_call variable
%type <lnode> cond_expr switch switch_case
%type <lnode> statements statement program def return expr4 expr31
%type <lnode> string expr0_1 expr25 expr27 expr23 expr22 while
%type <lnode> do for expr24 expr211 expr212 expr213 comma_expr for_expr
%type <lnode> semi_statement fun_decleration var_or_fun
%type <lnode> F_LOCAL_NAME fun_block block number F_INT F_FLOAT F_CONST0 F_CONST1

%type <number> F_SET_FALSE F_SET_TRUE F_THIS_PLAYER F_WRITE F_THIS_OBJECT
%type <number> assign

%type <string> F_FUNC_DECL F_IDENTIFIER F_STRING F_ARGUMENT F_OBJECT_NAME function_name
%%

all: { current_functions = 0; } program { prog = $2; };

program: program def
	{ if ($2 != 0) {
	    struct lnode_def *p = (struct lnode_def *)$2;

	    p->next=current_functions;
	    current_functions=p;
	    $$ = (struct lnode *)current_functions;
	  } else
	    $$ = $1;
        }
       |	 /* empty */ { $$ = 0; };

inheritance: F_INHERIT F_STRING ';'
	{
       	  int v;
	  char *file2=0;
	  struct inherit_list *list,*list2,*list3;
	  struct lnode_var_def *ob_vars;
	  struct lnode_def *pr;
	  OBJECT *ob;
	  extern char *check_file_name PROT((char *,int,int));
	  extern char *make_shared_string PROT((char *));
	  extern void load_ob_from_swap PROT((OBJECT *));
	  extern char *compile_file;
	  jmp_buf erc;
	  struct jmp_globals jmp;

	  if(!strcmp(compile_file,$2))
	    {
		free($2);
		yyerror("You cannont inherit oneself.");
		return 0;
	    }

	  if(new_inherit_list)
	     {
		   for(list=new_inherit_list;list;list=list->next)
			if(!strcmp($2,list->name))
			   {
				yyerror("Re-inheritance of '%s' is not allowed.",$2);
			        free($2);
    				return 0;
			   }
	    }

	 file2=$2;
	 set_jump_state(erc,&jmp);
	 if(!setjmp(error_recovery_context)) 
	    ob = find_object2(check_file_name($2,0,1),1);
	 else { ob=0;clear_jump_state(&jmp);}

  	 reset_jump_state(erc);
	 if(ob && !check_realm_access(0,ob))
	    {
		yyerror("You cannont inherit '%s'.",$2);
		free($2);
		return 0;
	    }
	 if (!ob)
	    {
		inherit_file = file2;
		if(new_inherit_list)
		  {
		    free_inherit_list(new_inherit_list,0);
		    new_inherit_list=0;
		  }
		return 0;
	    }
	 if(ob->flags & SWAPPED) load_ob_from_swap(ob);
	 copy_variables(ob);
	 if(ob->inherit)
	   {
		struct inherit_list *temp_inherit_list=0;
		for(list=ob->inherit;list;list=list->next)
		   {
			   int v;

			   list2=alloc_inherit(list->num_var);
			   list2->num_var=list->num_var;
			   list2->prog=list->prog;
			   add_prog_ref(list->prog);
			   list2->block=list->block;
			   list2->swap_num=list->swap_num;
			   list2->status=list->status;
			   list2->swap_size=list->swap_size;
			   list2->name=make_shared_string(list->name);

		   for(v=list2->num_var-1,ob_vars=list2->status;ob_vars;v--,ob_vars=ob_vars->next)
			list2->var_x_ref[ob_vars->num_var]=check_deklared(ob_vars->name); 

		  list2->next=temp_inherit_list;
		  temp_inherit_list=list2;
	       }

	    for(list=temp_inherit_list;temp_inherit_list;list=temp_inherit_list)
	      {
	   	temp_inherit_list=temp_inherit_list->next;
	   	list->next=new_inherit_list;
	   	new_inherit_list=list;
	      } 
  	   }

	list=alloc_inherit(ob->num_variables);
	list->num_var=ob->num_variables;
	list->prog=ob->prog;
	add_prog_ref(ob->prog);
	list->block=ob->block;
	list->swap_num=ob->swap_num;
	list->status=ob->status;
	list->swap_size=ob->swap_size;
	list->name=make_shared_string(ob->name);

	for(v=ob->num_variables-1,ob_vars=ob->status;ob_vars;v--,ob_vars=ob_vars->next)
	list->var_x_ref[ob_vars->num_var]=check_deklared(ob_vars->name); 

	list->next=new_inherit_list;
	new_inherit_list=list;
}

number:  F_INT | F_FLOAT | F_CONST0 | F_CONST1;

fun_type: type
	| F_VOID { fun_only=1; }
	| /* empty */ {fun_only=1;};

no_mask: F_NO_MASK {fun_only=1;}
	| /* empty */ ;

fun_decleration: private static {fun_only=0;} no_mask fun_type var_or_fun
	{ $$=$6; } ;

var_or_fun: name_list ';' { var_flags=0;$$=0;}
   | optional_astr F_IDENTIFIER
     { 
	  if(lookup_predef($2))
		yyerror("You cannot use the reserved word '%s' as a function name.\n",$2);

	  if(find_p_hash($2,compile_file))
		yyerror("You cannot delclare the function '%s' more than once.",$2);
     } 
    '(' argument ')' { num_arg=current_number_of_locals; } fun_block
	{ struct lnode_def *p;
	  int i;

	  if(!no_block)
           {
	     p = alloc_lnode_def(f_identifier,$2,$8,current_number_of_locals-num_arg);
	     p->num_arg = num_arg;
	     p->flags=function_flags;
	     $$ = (struct lnode *)p;
	   }
	  else $$=0;

          if (argument_name) free(argument_name);
          argument_name = 0; 
          free_all_local_names(); 
        } ;

fun_block: ';' {$$=0;no_block=1;}
	| block {$$=$1;no_block=0;};

def: F_PRIVATE ';' { object_flags|=PRIVATE;$$=0;}
   | F_GLOBAL ';' {object_flags|=GLOBAL;$$=0;}
   | F_NO_EXT_SAVE ';' {object_flags|=NO_EXT_SAVE;$$=0;}
   | F_NO_CLONE ';' {object_flags|=NO_CLONE;$$=0;}
   | F_RESET ';' {object_flags|=RESET;$$=0;}
   | inheritance { $$ = 0; } 
   | fun_decleration {$$=$1; } ;

argument: /* empty */
	| local_list ;

private: {function_flags=0;}
	| F_PRIVATE {function_flags=(unsigned char)L_PRIVATE;};

static: /* empty */
	| F_STATIC {function_flags|=L_STATIC;};

optional_astr: /* empty */
	       | '*' ;
				 
type: F_INT_DECL
    | F_MIXED 
    | F_FLOAT_DECL
    | F_OBJECT
    | F_STRING_DECL 
    | F_POINTER ;

name_list: new_name
	 | name_list ',' new_name;

new_name: optional_astr F_IDENTIFIER
	{
	   if(fun_only) yyerror("Invalid type to assigned to gobal variable '%s'.\n",$2);
	   if(!alloc_lnode_var_def(f_int, $2, variable_count++,function_flags))
	       yyerror("You cannont declare the global variable '%s' more then once.",$2); 
	}

block: '{' local_declarations statements '}'
        { $$ = (struct lnode *)alloc_lnode_block($3); };

local_declarations: /* empty */
	| local_declarations type local_list ';';

local_type: type
	  | /* empty */ ;

local_list: local_type optional_astr F_IDENTIFIER { add_local_name($3); }
	| local_list ',' local_type optional_astr F_IDENTIFIER { add_local_name($5); }
	| local_type optional_astr F_ARGUMENT { yyerror("You cannong declare the argument '%s' more than once.",$3); }
	| local_type optional_astr F_LOCAL_NAME { yyerror("You cannont declare the local variable '%s' more than once.",$3); };

statements: /* empty */ { $$ = 0; }
	| statement statements
	{ if ($1 == 0)
	      $$ = $2;
	  else
	      $$ = (struct lnode *)alloc_lnode_2(f_cons, $1, $2);
        }
	| error ';' { $$ = 0; };

statement: expr0_1 ';' { $$ = $1; }
	 | cond | switch | while | do | for 
	 | return ';' { $$ = $1; }
	 | block
	 | switch_case
	 | F_DEFAULT ':'
	{ $$=(struct lnode *)alloc_lnode_single(f_default);}
	 | F_BREAK ';'
	{ $$=(struct lnode *)alloc_lnode_single(f_break);}
	 | F_CONTINUE ';' 
	{ $$=(struct lnode *)alloc_lnode_single(f_continue);}

switch_case: F_CASE expr0_1 ':'
	{ $$=(struct lnode *)alloc_lnode_1(f_case,$2);}

switch: F_SWITCH '(' expr0_1 ')' statement
	{ $$=(struct lnode *)alloc_lnode_2(f_switch,$3,$5);}

while: F_WHILE '(' expr0_1 ')' semi_statement
	{ $$ = (struct lnode *)alloc_lnode_2(f_while, $3, $5); };

do: F_DO semi_statement F_WHILE '(' expr0_1 ')' ';'
	{ $$ = (struct lnode *)alloc_lnode_2(f_do, $2, $5); };

semi_statement: ';' { $$=0;} | statement {$$=$1;}
		
for: F_FOR '(' for_expr ';' expr0_1 ';' for_expr ')' semi_statement
	{ $$ = (struct lnode *)
	alloc_lnode_2(f_cons,$3,
	(struct lnode *) alloc_lnode_2(f_cons,
	(struct lnode *)alloc_lnode_3(f_for,$5,$7,$9),(struct lnode *)0)); 
	}

for_expr: /* EMPTY */
        { $$ = (struct lnode *)alloc_lnode_single(f_const1); }
        | comma_expr
        { $$=$1; }

comma_expr: expr0_1
	{ $$=$1; };
	| comma_expr ',' expr0_1
	{ 
	if(!$3) yyerror("Bad comma expression.");
	if($3->fun==f_comma)
	$$ = (struct lnode *)alloc_lnode_2(f_comma, $1, $3);
	else
	$$ = (struct lnode *)alloc_lnode_2(f_comma, $1,
	(struct lnode *)alloc_lnode_2(f_comma, $3, 0));
	}

cond_expr: expr1 { $$=$1; }
	| expr1 '?' expr0_1 ':' expr0_1 
	{ $$= (struct lnode *)alloc_lnode_3(f_if,$1,$3,$5);}

expr0_1: cond_expr
	| variable assign expr0_1
	{
	  switch($2)
	   {
		case F_AND_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_and_eq,$1,$3);
			break;
		case F_ASSIGN:
			$$ = (struct lnode *)alloc_lnode_2(f_assign,$1,$3);
			break;
		case F_OR_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_or_eq,$1,$3);
			break;
		case F_XOR_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_xor_eq,$1,$3);
			break;
		case F_LSH_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_lsh_eq,$1,$3);
			break;
		case F_RSH_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_rsh_eq,$1,$3);
			break;
		case F_ADD_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_add_eq,$1,$3);
			break;
		case F_SUB_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_sub_eq,$1,$3);
			break;
		case F_MULT_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_mult_eq,$1,$3);
			break;
		case F_MOD_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_mod_eq,$1,$3);
			break;
		case F_DIV_EQ:
			$$ = (struct lnode *)alloc_lnode_2(f_div_eq,$1,$3);
			break;
		defualt:
			yyerror("Unknown asignment type '%d'.",$2);
			$$=0;
	      }
	}
	| error assign expr0_1
	{
	  char *expr_name;

	  switch($2)
	   {

		case F_AND_EQ:
			expr_name="&=";
			break;
		case F_ASSIGN:
			expr_name="=";
			break;
		case F_OR_EQ:
			expr_name="|=";
			break;
		case F_XOR_EQ:
			expr_name="^=";
			break;
		case F_LSH_EQ:
			expr_name="<<=";
			break;
		case F_RSH_EQ:
			expr_name=">>=";
			break;
		case F_ADD_EQ:
			expr_name="+=";
			break;
		case F_SUB_EQ:
			expr_name="-=";
			break;
		case F_MULT_EQ:
			expr_name="*=";
			break;
		case F_MOD_EQ:
			expr_name="%=";
			break;
		case F_DIV_EQ:
			expr_name="/=";
	      }
	 yyerror("Bad expresion type to the left of '%s'.",expr_name);
         $$=0;
	};

assign: '=' { $$=F_ASSIGN; }
	| F_AND_EQ { $$=F_AND_EQ;}
	| F_DIV_EQ { $$=F_DIV_EQ;}
	| F_MOD_EQ { $$=F_MOD_EQ;}
	| F_MULT_EQ {$$=F_MULT_EQ;}
	| F_SUB_EQ  {$$=F_SUB_EQ;}
	| F_ADD_EQ  {$$=F_ADD_EQ;}
	| F_RSH_EQ  {$$=F_RSH_EQ;}
	| F_LSH_EQ  {$$=F_LSH_EQ;}
	| F_XOR_EQ  {$$=F_XOR_EQ;}
	| F_OR_EQ   {$$=F_OR_EQ;};

return: F_RETURN
	{ $$ = (struct lnode *)alloc_lnode_1(f_return, 0); }
	| F_RETURN expr0_1 
	{ $$ = (struct lnode *)alloc_lnode_1(f_return, $2); };

expr_list: /* empty */ { $$ = 0; }
	| expr0_1
	{ $$ = (struct lnode *)alloc_lnode_2(f_cons, $1, 0); }
	| expr0_1 ',' expr_list
	{ $$ = (struct lnode *)alloc_lnode_2(f_cons, $1, $3); };

expr1: expr2
	| expr2 F_LOR expr1
	{ $$ = (struct lnode *)alloc_lnode_2(f_lor, $1, $3); };

expr2: expr211
	| expr211 F_LAND expr2
	{ $$ = (struct lnode *)alloc_lnode_2(f_land, $1, $3); };

expr211: expr212
	| expr211 '|' expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_for, $1, $3); };

expr212: expr213
	| expr212 '^' expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_xor, $1, $3); };

expr213: expr22
	| expr213 '&' expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_and, $1, $3); };

expr22: expr23
	| expr24 F_EQ expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_eq, $1, $3); }
	| expr24 F_NE expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_ne, $1, $3); };

expr23: expr24
	| expr24 '>' expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_gt, $1, $3); }
	| expr24 F_GE expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_ge, $1, $3); }
	| expr24 '<' expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_lt, $1, $3); }
	| expr24 F_LE expr24
	{ $$ = (struct lnode *)alloc_lnode_2(f_le, $1, $3); };

expr24: expr25
	| expr24 F_LSH expr25
	{ $$ = (struct lnode *)alloc_lnode_2(f_lsh, $1, $3); }
	| expr24 F_RSH expr25
	{ $$ = (struct lnode *)alloc_lnode_2(f_rsh, $1, $3); };

expr25: expr27
	| expr25 '+' expr27
	{ $$ = (struct lnode *)alloc_lnode_2(f_add, $1, $3); }
	| expr25 '-' expr27
	{ $$ = (struct lnode *)alloc_lnode_2(f_subtract, $1, $3); };

expr27: expr3
	| expr27 '*' expr3
	{ $$ = (struct lnode *)alloc_lnode_2(f_multiply, $1, $3); }
	| expr27 '%' expr3
	{ $$ = (struct lnode *)alloc_lnode_2(f_mod, $1, $3); }
	| expr27 '/' expr3
	{ $$ = (struct lnode *)alloc_lnode_2(f_divide, $1, $3); };

expr3: expr31
	| F_INC variable
        { $$ = (struct lnode *)alloc_lnode_1(f_inc, $2); };
	| F_DEC variable
        { $$ = (struct lnode *)alloc_lnode_1(f_dec, $2); };
	| F_NOT expr3
	{ $$ = (struct lnode *)alloc_lnode_1(f_not, $2); };
	| '~' expr3
	{ $$ = (struct lnode *)alloc_lnode_1(f_compl, $2); };
	| '-' expr3
	{ $$ = (struct lnode *)alloc_lnode_1(f_negate, $2); };

expr31: expr4
	| variable F_INC
	{ $$ = (struct lnode *)alloc_lnode_1(f_post_inc, $1); };
	| variable F_DEC
	{ $$ = (struct lnode *)alloc_lnode_1(f_post_dec, $1); };

expr4:  function_call
	| variable
	| string 
	| number
	| '(' comma_expr ')' { $$ = $2; }
	| '(' '{' expr_list '}' ')'
	{ $$ = (struct lnode *)alloc_lnode_1(f_aggregate, $3); };

variable: F_IDENTIFIER
	{ $$ =(struct lnode *)alloc_lnode_number(f_identifier,check_deklared($1),0.0);
	  free($1);
	}
	| F_ARGUMENT { $$ = (struct lnode *)alloc_lnode_single(f_argument); }
	| F_LOCAL_NAME
	| expr4 '[' expr0_1 F_DOT_DOT expr0_1 ']'
	{ $$ = (struct lnode *)alloc_lnode_3(f_braced_extract, $1, $3, $5); }
	| expr4 '[' expr0_1 ']'
	{ $$ = (struct lnode *)alloc_lnode_2(f_subscript, $1, $3); };

string: F_STRING
	{ $$ = (struct lnode *)alloc_lnode_name(f_string, $1); };

function_call:
	function_name '(' expr_list ')'
	  { 
	    struct value *(*f)();
	    f=lookup_predef($1);
            if(f) $$ =(struct lnode *)alloc_lnode_1(f,$3);
	      else
		 $$ = (struct lnode *)alloc_lnode_funcall(f_function, $1, $3); 
	  }
	| expr4 F_ARROW function_name '(' expr_list ')'
	  {
	    struct lnode *p;
	    p = (struct lnode *)alloc_lnode_name(f_string, $3);
	    p = (struct lnode *)alloc_lnode_2(f_cons, p, $5);
	    p = (struct lnode *)alloc_lnode_2(f_cons, $1, p);
	    $$ = (struct lnode *)alloc_lnode_1(f_call_other, p);
	 };

function_name: F_IDENTIFIER
	| F_COLON_COLON F_IDENTIFIER
	{
	    char *p = xalloc(strlen($2) + 3);
	    strcpy(p, "::"); strcat(p, $2); free($2);
	    $$ = p;
	};
	|  F_OBJECT_NAME F_IDENTIFIER
	{
		char *p=xalloc(strlen($1)+strlen($2)+1);
		strcpy(p,$1);strcat(p,$2);
		free($1);free($2);
		$$ = p;
	};

cond: F_IF '(' expr0_1 ')' statement F_ELSE statement
	{ $$ = (struct lnode *)alloc_lnode_3(f_if, $3, $5, $7); }
	| F_IF '(' expr0_1 ')' statement
	{ $$ = (struct lnode *)alloc_lnode_3(f_if, $3, $5, 0); }
%%


/*VARARGS1*/
void yyerror(err,e1,e2,e3,e4,e5,e6,e7,e8,e9)
char *err;
int e1,e2,e3,e4,e5,e6,e7,e8,e9;
{
    extern int num_parse_error;
    char str[256];
    struct realms *r;
    extern char yytext[];
 
    if(strstr(err,"syntax error"))
	{
	   syntax_error++;
	   (void)sprintf(str,"%s on or before '%s'.",err,yytext);
	}
     else (void)sprintf(str,err,e1,e2,e3,e4,e5,e6,e7,e8,e9);


    if (num_parse_error > 5 || syntax_error>1)
	return;

    debug_message(extract_realm(current_file),"%s(%d) : %s\n",current_file,current_line,str);
    smart_log(current_file, current_line, str,0);
    if (num_parse_error == 0)
	save_error(str, current_file, current_line);
    num_parse_error++;
}

static int check_declared(str)
    char *str;
{
    struct lnode_var_def *p;


    for (p = prog_status; p; p = p->next) {
	int res;

	res=strcmp(str,p->name);
	if(!res) return p->num_var;
	if(res<0) return -1;
    }
    return -1;
}

static int check_deklared(str)
    char *str;
{
    int r;

    r = find_var_hash(str);
    if (r < 0) yyerror("Variable '%s' not declared!", str);
    return r;
}

void free_all_local_names()
{
    int i;

    for (i=0; i<current_number_of_locals; i++) {
	free(local_names[i]);
	local_names[i] = 0;
    }
    current_number_of_locals = 0;
}

void add_local_name(str)
    char *str;
{
	if(current_number_of_locals==cf.max_local)
	  yyerror("To many local variables defined.  Max: %d.",cf.max_local);
	local_names[current_number_of_locals++]=str;
}

static void copy_variables(from)
	struct object *from;
{
    struct lnode_var_def *p;

    if(!from->status) return;

    for(p=from->status;p;p=p->next)
     {
        if(p->flags & L_PRIVATE) continue;
	if(find_var_hash(p->name)!=-1)
	  {
		 yyerror("Duplicate variable '%s' in inherted object: %s.",p->name,from->name);
	     return;
	  }
        (void)alloc_lnode_var_def(f_int, string_copy(p->name), variable_count++,p->flags);
    }
}

void init_max_local()
{
  local_names=(char **)xalloc(sizeof (char *) * cf.max_local);

  memset((char *)local_names,'\0',sizeof (char *) * cf.max_local); 
}
