#include <stdio.h>
#include <string.h>
#include "config.h"
#include "lint.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "lang.h"
#include "error.h"
#include "lan_fun.h"
#include "s_hash.h"
#include "var_hash.h"
#include "p_hash.h"
#include "realms.h"

extern void add_message();
extern void fatal();
extern struct object_status *ob_status[];
extern int current_line;
int tot_alloc_lnode;

struct lnode_var_def *prog_status;

void free_lnode PROT((struct lnode *,int));

void free_sub_part PROT((struct lnode *, int));

static int tot_lnode_single;
struct lnode_single *alloc_lnode_single(fun)
    struct value *(*fun)();
{
    struct lnode_single *p;
    p=(struct lnode_single *)xalloc(sizeof (struct lnode_single));
    p->fun=fun;
    p->line=current_line+L_SINGLE;
	tot_lnode_single++;
    return p;
}

static struct lnode_number *lnode_number_list;
static int tot_lnode_number;
struct lnode_number *alloc_lnode_number(fun, number,n_float)
    struct value *(*fun)();
    int number;
    double n_float;
{
    struct lnode_number *p;

    p=(struct lnode_number *)xalloc(sizeof (struct lnode_number));
    tot_lnode_number++;
    p->fun = fun;
    p->number = number;
    p->n_float=n_float;
    p->line = current_line + L_NUMBER;
    return p;
}

/*
 * Strings are allocated from the shared string area.
 * Be carfule not to free() such an area through free().
 */
static struct lnode_name *lnode_name_list;
static int tot_lnode_name;
struct lnode_name *alloc_lnode_name(fun, name)
    char *name;
    struct value *(*fun)();
{
    struct lnode_name *p;
    char *spn;

    p=(struct lnode_name *)xalloc(sizeof (struct lnode_name));
    tot_lnode_name++;
    p->fun = fun;
    p->name = make_shared_string(name);
    free(name);
    p->line = current_line + L_NAME;
    return p;
}

static struct lnode_1 *lnode_1_list;
static int tot_lnode_1;
struct lnode_1 *alloc_lnode_1(fun, expr)
    struct lnode *expr;
    struct value *(*fun)();
{
    struct lnode_1 *p;

    p=(struct lnode_1 *)xalloc(sizeof (struct lnode_1));
    tot_lnode_1++;
    p->fun = fun;
    p->expr = expr;
    p->line = current_line + L_1;
    return p;
}

static struct lnode_2 *lnode_2_list;
static int tot_lnode_2;
struct lnode_2 *alloc_lnode_2(fun, expr1, expr2)
    struct lnode *expr1, *expr2;
    struct value *(*fun)();
{
    struct lnode_2 *p;

    p=(struct lnode_2 *)xalloc(sizeof (struct lnode_2));
	tot_lnode_2++;
    p->fun = fun;
    p->line = current_line + L_2;
    p->expr1 = expr1;
    p->expr2 = expr2;
    return p;
}

static int tot_lnode_3;
struct lnode_3 *alloc_lnode_3(fun, expr1, expr2, expr3)
    struct lnode *expr1, *expr2, *expr3;
    struct value *(*fun)();
{
    struct lnode_3 *p;

    p = (struct lnode_3 *)xalloc(sizeof (struct lnode_3));
    p->fun = fun;
    p->line = current_line + L_3;
    p->expr1 = expr1;
    p->expr2 = expr2;
    p->expr3 = expr3;
    tot_lnode_3++;
    return p;
}

static int tot_lnode_def;
struct lnode_def *alloc_lnode_def(fun, name, block, num_var)
    struct lnode *block;
    char *name;
    int num_var;
    struct value *(*fun)();
{
    struct lnode_def *p;
    extern char *compile_file;
    char *hash_name=make_hash_name(name,compile_file);

    p = (struct lnode_def *)xalloc(sizeof (struct lnode_def));
    p->fun = fun;
    p->line = current_line + L_DEF;
    p->name = make_shared_string(hash_name);
    free(hash_name);
    free(name);
    p->length = strlen(p->name);
    p->num_var =(unsigned short)num_var;
    p->block = block;
    p->next = 0;
    p->num_ref = 0;
    p->flags = 0;
    p->num_arg = 0;
    tot_lnode_def++;
    add_p_hash(p);
    return p;
}

static int tot_var_def;
int alloc_lnode_var_def(fun, name, num_var,flags)
    struct value *(*fun)();
    char *name;
    int num_var;
    int flags;
{
    struct lnode_var_def *p,*p2,*p3=0;
    extern char *compile_file;

    p = (struct lnode_var_def *)xalloc(sizeof (struct lnode_var_def));
    p->fun = fun;
    p->line = current_line + L_VAR_DEF;
    p->name = make_shared_string(name);
    p->num_var =(unsigned short)num_var;

    p->flags=(unsigned char)flags;

    add_var_hash(p);

    if(!prog_status)
      {
	prog_status=p;
	p->next=0;
      }
     else 
      {
       p->next=prog_status;
       prog_status=p;
      }

    free(name);
    tot_var_def++;
    return 1;
}

struct lnode_funcall *funcall_list=0;
static int tot_funcall;
struct lnode_funcall *alloc_lnode_funcall(fun, name, arg)
    struct lnode *arg;
    char *name;
    struct value *(*fun)();
{
    struct lnode_funcall *p;
    char *spn;

    p = (struct lnode_funcall *)xalloc(sizeof (struct lnode_funcall));
    p->fun = fun;
    p->line = current_line + L_FUNCALL;
    p->name = make_shared_string(name);
    free(name);
    p->expr = arg;
    p->next=funcall_list;
    funcall_list=p;
    p->fun_p=0;
    tot_funcall++;
    return p;
}

int assign_fun_p(ob)
	OBJECT *ob;
{
	register struct lnode_funcall *fun,*fun2;
        int ret=1;

     for(fun=funcall_list;fun;fun=fun2)
      {
	fun2=fun->next;
	fun->next=0;
	fun->fun_p=find_fun_p(fun->name,ob,1);
        if(!fun->fun_p)
         {
          debug_message(ob->realm,"%s:(%d) : Local function '%s' not defined.\n",
		ob->name,fun->line & ~L_MASK,fun->name);
	  ret=0;
         }
      }
     funcall_list=0;
     return ret;
}

/*
 * Statements in a block (list of statements) are stored with cons
 * node (lnode_2). This is very expensive, as one lnode_2 is required as
 * overhead for every statement. This function will take such a list
 * and move all statements in a block into one single memory vector, storing
 * them consecutively.
 */
int tot_lnode_block;
struct lnode_block *alloc_lnode_block(p)
    struct lnode *p;
{
    int size, num;
    char *block, *block2;
    struct lnode *l, *next;
    struct lnode_block *ret;

    if (!p) return 0;
    for (num = 0, size = 0, l=p; l; l = l->a2) 
	   {
		size += lnode_size[l->a1->line >> L_SHIFT];
		num++;
       }
    if (num == 1) 
	  {
		/* Only one statement. No need to make a block ! */
		struct lnode *stat;

		stat = p->a1;
		free_lnode(p,1);
		return (struct lnode_block *)stat;
        }
    block = xalloc(sizeof (int) + size + 100);
    block2 = block;
    for (l=p; l; l = next) 
	  {
		int s;

		s = lnode_size[l->a1->line >> L_SHIFT];
		memcpy(block2, (char *)l->a1, s);
		if((l->a1->line & L_MASK) == L_FUNCALL)
		  {
		     struct lnode_funcall *fun1,*fun2,*fun3;

		     fun1=(struct lnode_funcall *)l->a1;
		     if(fun1==funcall_list)
		        funcall_list=(struct lnode_funcall *)block2;
		     else for(fun2=funcall_list->next,fun3=funcall_list;fun2;
			      fun3=fun2,fun2=fun2->next)
				if(fun2==fun1)
			        {
			          fun3->next=(struct lnode_funcall *)block2;
				  break;
			        }
		  }
			
		free_lnode(l->a1, 0);	/* Node has been copied. */
		l->a1 = 0;
		block2 += s;
		next = l->a2;
		free_lnode(l, 1);
      }
    ret = (struct lnode_block *)xalloc(sizeof (struct lnode_block));
    ret->fun = f_block;
    ret->line = current_line + L_BLOCK;
    ret->block = block;
    ret->num_nodes = num;
    tot_lnode_block++;
    return ret;
}

void print_lnode_status()
{
    int tot =
	tot_lnode_single * sizeof (struct lnode_single) +
	tot_lnode_number * sizeof (struct lnode_number) +
	tot_lnode_name * sizeof (struct lnode_name) +
	tot_lnode_1 * sizeof (struct lnode_1) +
	tot_lnode_2 * sizeof (struct lnode_2) +
	tot_lnode_3 * sizeof (struct lnode_3) +
	tot_lnode_def * sizeof (struct lnode_def) +
	tot_var_def * sizeof (struct lnode_var_def) +
	tot_funcall * sizeof (struct lnode_funcall) +
	tot_lnode_block * sizeof (struct lnode_block);
    add_message("lnode:\n");
    add_message("   single:     %5d   %8d\n", tot_lnode_single,
		tot_lnode_single * sizeof (struct lnode_single));
    add_message("   numbers:    %5d   %8d\n", tot_lnode_number,
		tot_lnode_number * sizeof (struct lnode_number));
    add_message("   names:      %5d   %8d\n", tot_lnode_name,
		tot_lnode_name * sizeof (struct lnode_name));
    add_message("   1 size expr:%5d   %8d\n", tot_lnode_1,
		tot_lnode_1 * sizeof (struct lnode_1));
    add_message("   2 size expr:%5d   %8d\n", tot_lnode_2,
		tot_lnode_2 * sizeof (struct lnode_2));
    add_message("   3 size expr:%5d   %8d\n", tot_lnode_3,
		tot_lnode_3 * sizeof (struct lnode_3));
    add_message("   functions:  %5d   %8d\n", tot_lnode_def,
		tot_lnode_def * sizeof (struct lnode_def));
    add_message("   glob vars:  %5d   %8d\n", tot_var_def,
		tot_var_def * sizeof (struct lnode_var_def));
    add_message("   fun calls:  %5d   %8d\n", tot_funcall,
		tot_funcall * sizeof (struct lnode_funcall));
    add_message("   blocks:     %5d   %8d\n", tot_lnode_block,
		tot_lnode_block * sizeof (struct lnode_block));
    add_message("   Total bytes       %10d\n", tot);
}

int lnode_size[(L_MAX >> L_SHIFT) + 1];

/*
 * Setup fast access to the size of an lnode.
 */

void compute_lnode_size()
{
    lnode_size[L_SINGLE >> L_SHIFT] = sizeof (struct lnode_single);
    lnode_size[L_NUMBER >> L_SHIFT] = sizeof (struct lnode_number);
    lnode_size[L_NAME >> L_SHIFT] = sizeof (struct lnode_name);
    lnode_size[L_1 >> L_SHIFT] = sizeof (struct lnode_1);
    lnode_size[L_2 >> L_SHIFT] = sizeof (struct lnode_2);
    lnode_size[L_3 >> L_SHIFT] = sizeof (struct lnode_3);
    lnode_size[L_DEF >> L_SHIFT] = sizeof (struct lnode_def);
    lnode_size[L_VAR_DEF >> L_SHIFT] = sizeof (struct lnode_var_def);
    lnode_size[L_FUNCALL >> L_SHIFT] = sizeof (struct lnode_funcall);
    lnode_size[L_BLOCK >> L_SHIFT] = sizeof (struct lnode_block);
}

void free_lnode(p, update_count)
    struct lnode *p;
    int update_count;
{
    update_count=1;
    switch(p->line & L_MASK) 
	  {
    	case L_SINGLE:
			if (update_count) tot_lnode_single--;
			break;
    	case L_NUMBER:
			if(update_count) tot_lnode_number--;
			break;
	    case L_NAME:
			if(update_count) tot_lnode_name--;
			break;
 	    case L_1:
			if(update_count) tot_lnode_1--;
			break;
	    case L_2:
			if(update_count) tot_lnode_2--;
			break;
	    case L_3:
			if (update_count) tot_lnode_3--;
			break;
	    case L_DEF:
			if (update_count) tot_lnode_def--;
			del_p_hash(((struct lnode_def *)p)->name);
			break;
	    case L_VAR_DEF:
			if (update_count) tot_var_def--;
			break;
	    case L_FUNCALL:
			if (update_count) tot_funcall--;
			break;
	    case L_BLOCK:
			if (update_count) tot_lnode_block--;
			break;
	    default:
			fatal("Line: %d - Bad type in lnode_size(): 0x%x\n",p->line & ~L_MASK, p->line & L_MASK);
      }
    free((char *)p);
}

void add_prog_ref(p)
    struct lnode_def *p;
{
    p->num_ref++;
}

/*
 * The program is stored at one of two places. Either a tree pointed
 * to by ob->prog (first argument), or a consecutive area pointed
 * to by 'block'. These two different trees has to be free'd in different
 * ways. It is load_ob_from_swap() that loads a prog into the ob->block
 * instead of ob->prog. ob->block will also contain all variable definitions.
 */
int free_prog(p, block, size)
    struct lnode_def *p;
    int size;
    char *block;
{
    extern int total_num_prog_blocks, total_prog_block_size;
	extern void remove_swap_file();

    if(!p) 
      {
	debug_message(0,"Tried to free a non-program\n");
	return 0;
      }
    p->num_ref--;
    if (p->num_ref > 0) return 0;
    if (block) 
	  {
		free((char *)block); /* FREE BLOCK */
		total_prog_block_size -= size;
		total_num_prog_blocks--;
		remove_swap_file();
      }
     else free_sub_part((struct lnode *)p, 1);
    return 1;
}

/*
 * Free the space of one node, and all linkes from it.
 * When a L_BLOCK is freed, don't call free_lnode() with the nodes
 * in the top of the block, because these are allocated in one big
 * chunk.
 */
void free_sub_part(p, do_free)
    struct lnode *p;
    int do_free;
{
    extern int tot_alloc_strings;
    if (!p) return;
    switch(p->line & L_MASK) {
    case L_SINGLE:
	p->fun=0;
	break;
    case L_NUMBER:
	p->fun=0;
	break;
    case L_NAME:
	p->fun=0;
	free_string(((struct lnode_name *)p)->name,"free_sub_part: L_NAME");
	break;
    case L_1:
	free_sub_part(((struct lnode_1 *)p)->expr, 1);
	((struct lnode_1 *)p)->expr = 0;
	p->fun=0;
	break;
    case L_2:
	free_sub_part(((struct lnode_2 *)p)->expr1, 1);
	((struct lnode_2 *)p)->expr1 = 0;
	free_sub_part(((struct lnode_2 *)p)->expr2, 1);
	((struct lnode_2 *)p)->expr2 = 0;
	p->fun=0;
	break;
    case L_3:
	free_sub_part(((struct lnode_3 *)p)->expr1, 1);
	((struct lnode_3 *)p)->expr1 = 0;
	free_sub_part(((struct lnode_3 *)p)->expr2, 1);
	((struct lnode_3 *)p)->expr2 = 0;
	free_sub_part(((struct lnode_3 *)p)->expr3, 1);
	((struct lnode_3 *)p)->expr3 = 0;
	p->fun=0;
	break;
    case L_DEF:
    {
	struct lnode_def *dp = (struct lnode_def *)p;

	   del_p_hash(((struct lnode_def *)dp)->name);
	   free_string(dp->name,"free_sub_part: L_DEF");
	   dp->name = 0;
	free_sub_part(dp->block, 1);
	dp->block = 0;
	if (dp->next) {
	    free_sub_part((struct lnode *)dp->next, 1);
	    dp->next = 0;
	}
	p->fun=0;
	if(do_free) /* we don't want to do this in free_lnode as both
		       clean up the hash table. I should probably
		       change this later, for better efficancy in
		       house cleaning. */
	  {
	     tot_lnode_def--;
	     free((char *)p);
	     return;
	   }
	break;
    }
    case L_VAR_DEF:
    {
	struct lnode_var_def *lp;
	lp = (struct lnode_var_def *)p;
	if (lp->next)
	    free_sub_part((struct lnode *)lp->next, 1);
	free_string(lp->name,"free_sub_part: L_VAR_DEF");
	lp->name = 0;
	p->fun=0;
	break;
    }
    case L_FUNCALL:

	free_string(((struct lnode_funcall *)p)->name,"free_sub_part: L_FUNCTION");
	((struct lnode_funcall *)p)->name = 0;
	free_fun_p(((struct lnode_funcall *)p)->fun_p);
	free_sub_part(((struct lnode_funcall *)p)->expr, 1);
	((struct lnode_funcall *)p)->expr = 0;
	p->fun=0;
	break;
    case L_BLOCK:
    {
	struct lnode_block *lb = (struct lnode_block *)p;
	char *block = lb->block;
	int i;
	
	for (i=0; i < lb->num_nodes; i++) {
	    free_sub_part((struct lnode *)block, 0);
	    block += lnode_size[((struct lnode *)block)->line >> L_SHIFT];
	}
	p->fun=0;
	free((char *)lb->block); /* FREE BLOCK */
	break;
    }
    default:
	fatal("Line: %d - Bad type in free_sub_part(): 0x%x\n",p->line & ~L_MASK, p->line & L_MASK);
    }
    if (do_free) free_lnode(p, 1);
}

#if 0

#define NUM_LNODE_NUMBER_CONST		20
static struct lnode_number lnode_number_const[NUM_LNODE_NUMBER_CONST];
static struct lnode_number lnode_number_const_id[NUM_LNODE_NUMBER_CONST];
static struct lnode_number lnode_number_const_loc[NUM_LNODE_NUMBER_CONST];
static int num_number_match;

static struct value *(*lnode_func_list[])() =
{
    f_enable_commands, f_this_object, f_this_player, f_time, f_users,
    f_return,
};
#define NUM_FUNCS	(sizeof lnode_func_list / sizeof lnode_func_list[0])
static struct lnode_1 funcalls[NUM_FUNCS];
static int num_funcall_match;

init_constant_lnodes() {
    int i, num;

    for (i=0; i < NUM_FUNCS; i++) {
	funcalls[i].line = L_CONSTANT;
	funcalls[i].fun = lnode_func_list[i];
	funcalls[i].expr = 0;
    }
    for (i=0; i < NUM_LNODE_NUMBER_CONST; i++) {
	lnode_number_const[i].line = L_CONSTANT;
	lnode_number_const[i].fun = f_int;
	lnode_number_const[i].number = i;
	lnode_number_const_id[i].line = L_CONSTANT;
	lnode_number_const_id[i].fun = f_identifier;
	lnode_number_const_id[i].number = i;
	lnode_number_const_loc[i].line = L_CONSTANT;
	lnode_number_const_loc[i].fun = f_local_name;
	lnode_number_const_loc[i].number = i;
    }
}

struct lnode_single lnode_const0 = { L_CONSTANT, f_const0 };
struct lnode_single lnode_const1 = { L_CONSTANT, f_const1 };
int num_lnode_single;

int match_and_replace_lnode(match, destp)
    struct lnode *match, **destp;
{
    struct lnode_number *p = (struct lnode_number *)match;
    struct lnode_1 *p1 = (struct lnode_1 *)match;

    if (match->fun == f_const0) {
	*destp = (struct lnode *)&lnode_const0;
	num_lnode_single++;
	return 1;
    }
    if (match->fun == f_const1) {
	*destp = (struct lnode *)&lnode_const1;
	num_lnode_single++;
	return 1;
    }
    if ((p1->line & L_MASK) == L_1 && p1->expr == 0) {
	int i;

	for (i=0; i < NUM_FUNCS; i++) {
	    if (p1->fun != lnode_func_list[i])
		continue;
	    *destp = (struct lnode *)&funcalls[i];
	    num_funcall_match++;
	    return 1;
	}
	return 0;
    }
    if ((p->line & L_MASK) == L_NUMBER && p->fun == f_int) {
	if (p->number < 0 || p->number >= NUM_LNODE_NUMBER_CONST)
	    return 0;
	*destp = (struct lnode *)&lnode_number_const[p->number];
	num_number_match++;
	return 1;
    }
    if ((p->line & L_MASK) == L_NUMBER && p->fun == f_identifier) {
	if (p->number < 0 || p->number >= NUM_LNODE_NUMBER_CONST)
	    return 0;
	*destp = (struct lnode *)&lnode_number_const_id[p->number];
	num_number_match++;
	return 1;
    }
    if ((p->line & L_MASK) == L_NUMBER && p->fun == f_local_name) {
	if (p->number < 0 || p->number >= NUM_LNODE_NUMBER_CONST)
	    return 0;
	*destp = (struct lnode *)&lnode_number_const_loc[p->number];
	num_number_match++;
	return 1;
    }
    return 0;
}

void status_lnodes_matched() {
    add_message("Saved lnode_numbers: %6d (%7d bytes)\n", num_number_match,
		num_number_match * sizeof (struct lnode_number) -
		sizeof lnode_number_const_loc -
		sizeof lnode_number_const_id -
		sizeof lnode_number_const);
    add_message("Saved lnode_1      : %6d (%7d bytes)\n", num_funcall_match,
		sizeof (struct lnode_1) * (num_funcall_match - NUM_FUNCS));
    add_message("Saved lnode_single : %6d (%7d bytes)\n", num_lnode_single,
		sizeof (struct lnode_single) * (num_lnode_single - 2));
}

#endif

int tot_alloced_inherit;

struct inherit_list *alloc_inherit(num_var)
   int num_var;
{
   struct inherit_list *list;
   int size;

   size=sizeof (struct inherit_list)+num_var* sizeof (int);
   list=(struct inherit_list *)xalloc(size);
   list->name=0;
   list->prog=0;
   list->num_var=0;
   list->swap_num=0;
   list->status=0;
   list->heart_beat=0;
   list->catch_tell=0;
   list->block=0;
   list->swap_size=0;
   list->next=0;
   list->var_x_ref[0]=0;
   tot_alloced_inherit++;
   return list;
}

void free_inherit(list)
    struct inherit_list *list;
{
     free_string(list->name,"free_inherit");
     list->name=0;
     list->prog=0;
     list->block=0;
     list->status=0;
     list->swap_size=0;
     list->swap_num=0;
     list->next=0;
     list->num_var=0;
     list->heart_beat=0;
     list->catch_tell=0;
     free((char *)list);
     tot_alloced_inherit--;
}

void free_inherit_list(list,clone_flag)
	struct inherit_list *list;
	int clone_flag;
{
    if(!list) return;
	for(;list;)
	  {
		struct inherit_list *tmp;
	
		tmp=list;
		list=list->next;

		if(!clone_flag) tmp->next=0;

 	        if(free_prog((struct lnode_def *)tmp->prog,tmp->block,tmp->swap_size))
	          {
		 	if(tmp->swap_num==-1 && tmp->status)
			   free_sub_part((struct lnode *)tmp->status,1);
		  }

		if(!clone_flag) free_inherit(tmp);
		tmp=0;
          }
}
