/*
 * This is a generic object. Clone a copy, and
 * setup local values.
 */

/*
 * If you are going to copy this file, in the purpose of changing
 * it a little to your own need, beware:
 *
 * First try one of the following:
 *
 * 1. Do clone_object(), and then configur it. This object is specially
 *    prepared for configuration.
 *
 * 2. If you still is not pleased with that, create a new empty
 *    object, and make an inheritance of this objet on the first line.
 *    This will automatically copy all variables and functions from the
 *    original object. Then, add the functions you want to change. The
 *    original function can still be accessed with '::' prepended on the name.
 *
 * The maintainer of this mud might become sad with you if you fail
 * to do any of the above. Ask other wizards if you are doubtful.
 *
 * The reason of this, is that the above saves a lot of memory.
 */

#include "ids.h"

global;


#define BREAK	1	/* can it break                */
#define GET_OUT	2	/* can you get items from it   */
#define PUT_IN	3	/* can you put items in it     */
#define PICK_UP	4	/* can you pick it up          */
#define DROP	5	/* can it be dropped           */
#define DEST	6	/* will it destruct if dropped */
#define RAN_VAL	7	/* does this object have a random value */
#define INV	8	/* is the inventory visable    */
#define PREV_IN 9	/* can we go into a container  */


int status_flag;

float value, weight, max_weight;

string cat_file, read_msg, short_desc, long_desc;

static float curr_weight;	/* this value is not saved     */



short() { return short_desc; }

long() { return long_desc; }


  /* If there is a message to be read add the verb 'read' to the living
     object calling this function.
  */

init()
{
    if (read_msg || cat_file) add_action("read","read");
}

  /* If there is a message on the object that can be read this function
     will show the message to the object using the 'read' verb defined
     in the init statement.
  */

read(str) 
{
    if (!id(str)) return 0;

    if(read_msg) write(read_msg);
      else cat(cat_file);

    return 1;
}

 /* this function is called when the object is dropped.
    If it is can be, test to see if the object should self destruct.
 */

drop(silently)
{
  if(test_bit(status_flag,DEST)) 
    {
  	if (!silently) write ("The "+ids[0]+" vanishes.\n");
  	destruct(this_object());
    }
}

  /* add weight to this object when puting something inside it.
     if max_weight is set, test to see if the object can hold the
     new weight.  If it can't test to see if it can break or not.
     then return 0.  Else return 1 to indicate that the weight has been
     added to the object.
  */

add_weight(n)
{
  if(max_weight && n+curr_weight>max_weight) 
     {
        if(test_bit(status_flag,BREAK)) 
	  {
	     do_break_msg();
	     destruct(this_object());
	     return 0;
          } 
        return 0;
     }
  curr_weight+=0.0+n;
  return 1;
}

  /* this is done in a separate function so that for special items that
     inherit this object can modify it.
  */

static do_break_msg()
{
	write(ids[0]+" breaks from the stress.\n");
	say(this_player()->query_name()+" breaks the "+ids[0]+".\n");
}

  /* does this item have an inventory that you can see. */

set_cat_file(f) 
{
  if(file_size(f)<0) return;

  cat_file=f;
}

set_visable_inventory(n)
{
  if(n) status_flag=set_bit(status_flag,INV);
    else status_flag=clear_bit(status_flag,INV);
}

  /* is there a limit to what this item can hold. */

set_max_weight(n)
{
  if(n>=0) max_weight=0.0+n;
}

   /* set the short description of this item. */

set_short(str) { short_desc = str; }

  /* set the long description of this item. */

set_long(str) { long_desc = str; }

  /* what the value of this object is, if any. */

set_value(v) 
{
  if(v>=0) value = 0.0+v;  /* make sure this is a float value. */ 
}

set_random_value(n)
{
  if(n) status_flag=set_bit(status_flag,RAN_VAL);
   else status_flag=clear_bit(status_flag,RAN_VAL);
}

 /* set the weight of this object. */

set_weight(w) 
{
  if(w>=0) weight =0.0+w;  /* make sure this is af float value. */ 
}

  /* set the message that can be read from this object. */

set_msg(str) { read_msg = str; }

  /* if the object can be put into a container. */

set_prevent_insert(n)
{
  if(n) status_flag=set_bit(status_flag,PREV_IN);
    else status_flag=clear_bit(status_flag,PREV_IN);
}

  /* if the object can be gotten or not. */

set_can_get_out(n)
{
  if(n) status_flag=set_bit(status_flag,GET_OUT);
    else status_flag=clear_bit(status_flag,GET_OUT);
}

  /* if this object can be broken. */

set_can_break(n)
{
  if(n) status_flag=set_bit(status_flag,BREAK);
   else status_flag=clear_bit(status_flag,BREAK);
}

 /* can we put items in this object. */

set_can_put_in(n)
{
  if(n) status_flag=set_bit(status_flag,PUT_IN);
   else status_flag=clear_bit(status_flag,PUT_IN);
}

  /* can this item be picked up. */

set_can_pick_up(n)
{
  if(n) status_flag=set_bit(status_flag,PICK_UP);
   else clear_bit(status_flag,PICK_UP);
}

 /* can this item be dropped. */

set_can_drop(n) 
{
  if(n) status_flag=set_bit(status_flag,DROP);
   else status_flag=clear_bit(status_flag,DROP);
}

 /* will this item destruct if droped. */

set_destruct(n) 
{
  if(n) status_flag=set_bit(status_flag,DEST);
    else status_flag=clear_bit(status_flag,DEST);
}

  /* info functions on this object. */

query_destruct() { return test_bit(status_flag,DEST); }

query_can_pick_up() { return test_bit(status_flag,PICK_UP); }

query_can_drop() { return test_bit(status_flag,DROP); }

query_can_put_in() { return test_bit(status_flag,PUT_IN); }

query_can_get_out() { return test_bit(status_flag,GET_OUT); }

query_can_break() { return test_bit(status_flag,BREAK); }

query_prevent_insert() { return test_bit(status_flag,PREV_IN); }

query_visable_inventory() { return test_bit(status_flag,INV); }

query_weight() { return weight+curr_weight; }

query_value() 
{
   if(!test_bit(status_flag,RAN_VAL)) return value;

   return random(value); 
}

query_random_value() { return test_bit(status_flag,RAN_VAL); }

query_read_msg() { return read_msg; }

query_name() 
{
  if(!pointerp(ids)) return "";

  return ids[0]; 
}

query_cat_file() { return cat_file; }

status()
{
   write("Object Status for: "+file_name(this_object())+"\n");
   write("---------------------------------\n");
   write("Short   : "+short_desc+".\n\n");
   write("Long    : "+long_desc+"\n");
   write("Msg     : "+read_msg+"\n");
   write("Cat file: "+cat_file+"\n");
   write("IDS     : "+ids+"\n");
   write("Value   : "+value+(test_bit(status_flag,RAN_VAL)?" (random)":"")+"\n");
   write("Weight  : "+weight+"  ("+(curr_weight+weight)+")\n");
   write("Max/Curr: "+max_weight+"  ("+curr_weight+")\n\n");
   write("Flags   : ");

   if(test_bit(status_flag,DEST)) write("DEST, ");
   if(test_bit(status_flag,BREAK)) write("BREAK, ");
   if(test_bit(status_flag,PICK_UP)) write("PICK_UP, ");
   if(test_bit(status_flag,DROP)) write("DROP, ");
   if(test_bit(status_flag,GET_OUT)) write("GET_OUT, ");
   if(test_bit(status_flag,PUT_IN)) write("PUT_IN, ");
   if(test_bit(status_flag,INV)) write("INV, ");
   if(test_bit(status_flag,PREV_IN)) write("PREV_IN");
   write("\n");
}
