#include "object.h"
#include "soul.h"
#include "log.h"
#include "levels.h"
#include "wizard.h"
#include "race_class.h"

#define DIR_ARRAY ({"east","west","north","south","up","down","northeast","northwest","southeast","southwest","left","right","exit","leave","enter","in","out"})
#define MCP "r1/closed/mcp"

#define MAX_FOOD 10
#define MAX_FLUID 10
#define MAX_DRUNK 10

/*
 * If you are going to copy this file, in the purpose of changing
 * it a little to your own need, beware:
 *
 * First try one of the following:
 *
 * 1. Do clone_object(), and then configur it. This object is specially
 *    prepared for configuration.
 *
 * 2. If you still is not pleased with that, create a new empty
 *    object, and make an inheritance of this objet on the first line.
 *    This will automatically copy all variables and functions from the
 *    original object. Then, add the functions you want to change. The
 *    original function can still be accessed with '::' prepended on the name.
 *
 * The maintainer of this mud might become sad with you if you fail
 * to do any of the above. Ask other wizards if you are doubtful.
 *
 * The reason of this, is that the above saves a lot of memory.
 */

/*
 * Include this file in objects that "lives".
 * The following variables are defined here:
 *
 */
pointer armour_classes;
pointer armour,weapons,height,quests;
object attacker;
pointer levels,classes,race_min_max,class_min_max;
pointer skill_names,skill_levels;
int num_attacks;
pointer Wis;
string race;
pointer exp;
int can_talk;
int hunger,thirst,drunk;
int hit_points_die;
int local_weight;
int num_hands,hands_used;
int magical;
int playing_time;
int use_light;
int time_to_heal;	/* Count down variable. */
int credits;		/* Amount of credits on player. */
string name;		/* Name of object. */
string msgin, msgout;	/* Messages when entering or leaving a room. */
string whimpy_dir;
int wiz_tester;
int brief;	/* Flags. */
int wiz_level;		/* Level of monster. */
int hit_points;		/* Number of hit points of monster. */
string mmsgout;		/* Message when leaving magically. */
string mmsgin;		/* Message when arriving magically. */
int ghost;		/* Used for monsters that can leave a ghost. */
static string cap_name;	/* Capital version of "name". */
int spell_points;	/* Current spell points. Same max as max_hp. */
int age;		/* Number of heart beats of this character. */
int is_invis;		/* True when player is invisible */
int whimpy;		/* Automatically run when low on HP */
int dead;		/* Are we alive or dead? */
/*
 * All characters have an aligment, depending on how good or chaotic
 * they are.
 * This value is updated when killing other players.
 */
string alignment;
int gender;	/* 0 means neuter ("it"), 1 male ("he"),  2 female ("she") */
/*
 * Character stat variables.
 */
int Str, Int, Con, Dex;

/*
 * The following routines are defined for usage:
 * stop_fight		Stop a fight. Good for scroll of taming etc.
 * hit_player		Called when fighting.
 * transfer_all_to:	Transfer all objects to dest.
 * query_name:		Gives the name to external objects.
 * attacked_by		Tells us who are attacking us.
 * show_stats		Dump local status.
 * stop_wielding	Called when we drop a weapon.
 * stop_wearing		Called when we drop an armour.
 * query_level		Give our level to external objects.
 * attack		Should be called from heart_beat. Will maintain attack.
 * query_attack
 * wield		Called by weapons.
 * wear			Called by armour.
 * heal_self		Enable wizards to heal this object.
 * attack_object	Called when starting to attack an object.
 * test_if_any_here	For monsters. Call this one if you suspect no enemy
 *			is here any more.
 *			Return 1 if anyone there, 0 if none.
 * force_us		Force us to do a command.
 * query_spell_points	Return how much spell points the character has.
 */

init_arrays()
{
  if(!pointerp(Con)) Con=allocate(6);
  if(!pointerp(Int)) Int=allocate(6);
  if(!pointerp(Str)) Str=allocate(6);
  if(!pointerp(Dex)) Dex=allocate(4);
  if(!pointerp(Wis)) Wis=allocate(3);
  if(!pointerp(exp)) exp=allocate(3);
  if(!pointerp(classes)) classes=allocate(3);
  if(!pointerp(levels)) levels=allocate(3);
  if(!pointerp(race_min_max)) race_min_max=({({0,0}),({0,0}),({0,0}),({0,0})});
  if(!pointerp(class_min_max)) class_min_max=({({0,0}),({0,0}),({0,0}),({0,0}),({0,0})});
  if(!pointerp(hit_points_die)) hit_points_die=allocate(3);
  if(!pointerp(hit_points)) hit_points=allocate(2);
  if(!pointerp(spell_points)) spell_points=allocate(2);
  if(!pointerp(armour_classes)) armour_classes=({10,10,10});
  if(!pointerp(local_weight)) local_weight=allocate(2);
}

death(dam)
{
	object corpse;

	/* We died ! */
	
	dead = 1;

	corpse = clone_object("r1/obj/corpse");
	corpse->set_name(name);

 	if(this_player() && this_player()!=this_object() &&
           this_player()->query_name())
                 corpse->set_killer(this_player()->query_name());

	transfer_all_to(corpse);

	move_object(corpse, environment(this_object()));

	if(this_object()->second_life()) return dam;

        set_heart_beat(0);
        destruct(this_object());
        return dam;
}

transfer_all_to(dest)
{
    int s,i;
    pointer inv;
    object ob;

    if(!dest->query_can_put_in()) return;

    inv = all_inventory(this_object());

    for(s=sizeof(inv);i<s;i++)
    {
        /* make sure you have a valid object, call drop, then
           check again incase it was destructed.
        */

	if(inv[i] && inv[i]->query_can_drop() && inv[i]) 
	  {
	     add_weight(-inv[i]->query_weight());

	     if(inv[i]) move_object(inv[i], dest);
	  }
    }
}

query_name() { return cap_name; }

query_alignment() { return alignment; }

set_alignment(al) 
{
  if(numberp(al))
     if(al>=0 && al<sizeof(ALIGNMENT_ARRAY))
	  return alignment=ALIGNMENT_ARRAY[al];
       else return 0;

  if(search_array(ALIGNMENT_ARRAY,al)!=-1)
	return alignment=al;
      
  return 0;
 }

wield_weapon(wep,silently) 
{
int hands;
object w;

    if(hands_used==num_hands)
      {
	 if(!silently) write("You cannot wield anything more!\n");
	 return 1;
      }
    
    w=present(wep,this_object());

    if(!w) w=present(w,environment());

    if(!w)
      {
        if(!silently) write("There is no '"+wep+"' here.\n");
	return 1;
      }

    if(w->query_wielded())
      {
         if(!silently) write("You are already wielding that!\n");
         return 1;
      }
	 
    if(hands_used+(hands=w->query_num_hands())>num_hands) 
      {
         if(!silently) write("That is more than you can presently handle.\n");
         return 1;
      }

    hands_used+=hands;

    if(!weapons) weapons=({w});
	else weapons+=({w});

    if(!silently) MCP->do_action_msg(w,"wields");

    return 1;
}

sheith_weapon(wp,silently) 
{
int i;
object wep;

    if(dead) silently=1;
   
    if (!weapons) return;

    wep=present("wielded "+wp,this_object());

    if(!wep)
      if(!silently)
        {
           write("You are not wielding that weapon.\n");
	   return 1;
        }
     else return 1;

    i=search_array(weapons,wep);

    wep->sheith(silently);
    weapons=del_element(weapons,i);

   if(!silently) do_action_msg(wep,"sheiths");

   return 1;
}

wear(a,silently)
{
    object armr;
    int ac;

    armr=present(a,this_object());

    if(!armr) armr=present(a,environment());

    if(!armr)
     {
       if(!silently) write("There is no '"+a+"' here.\n");
       return 1;
     }

    if(a->query_worn())
     {
       if(!silently) write("You are already wearing that!\n");
       return 1;
     }

    if(present("worn "+a->query_type(),this_object()))
      {
        if(!silently) write("You must first remove what you already have on.\n");
	return 1;
      }

    if(a->query_armour())
      {
         armour_classes[0]+=(ac=a->query_ac());

         if(a->query_type()!="shield")
           {
        	armour_classes[1]+=ac;
		armour_classes[2]+=ac;
           }
       }
     else if(!a->query_clothes()) 
      {
	if(!silently) write("You cannot wear that!\n");
	return 1;
      }
    
    if(silently) return 1;

    MCP->do_action_msg(a,"wears");
    return 1;
}

remove(ar,silently) 
{
string check_for_a;
object armr;
int ac;

    if(dead) silently=1;

    armr=present("worn "+ar,this_object());

    if(!armr)
      if(!silently)
       {
          write("You are not wearing it.\n");
	  return 1;
       }
     else return 1;

    if(armr->query_armour())
     {
        armour_classes[0]-=(ac=armr->query_ac());

        if(armr->query_type()!="shield")
          {
            armour_classes[1]-=ac;
            armour_classes[1]-=ac;
          }
     }

    armr->remove();
    
    if(!silently) MCP->do_action_msg(armr,"removes");

    return 1;
}

query_level(cls)
{
int i;

    if(!cls)
     {
       if(levels[0]>levels[1]) return levels[0];

       return levels[1];
     }

    if(cls==-1) return levels;

    if((i=search_array(classes,cls))==-1) return 0;

    return levels[i];
}

add_magic_armour(num) 
{ 
   armour_classes[0]+=num;
   armour_classes[1]+=num;
   armour_classes[2]+=num;
}

add_weight(w,n) {

    if (!n && (w + local_weight[1]) > (Str[3]+(Str[4]/2)) ) return 0;

    if(!n) local_weight[1] +=0.0+w;
	else local_weight[0]+=0.0+w;

    return 1;
}

heal_self(h) 
{
    if (h <= 0) return;

  if(this_player()!=this_object() && test_player(this_player()))
      log_file("HEAL_SELF",this_player()->query_real_name()+ " healed "+name+" for "+h+" points on:"+ctime(time())+"\n");

    if(hit_points[0]<hit_points[1])
      {
         hit_points[0] += h;
         if (hit_points[0] > hit_points[1]) hit_points[0] = hit_points[1];
      }

    if(!magical) return;

    if(spell_points[0]<spell_points[1])
     {
         spell_points[0] += h;
         if (spell_points[0] > spell_points[1]) spell_points[0] = spell_points[1];
     }
}

adj_spell_points(h) 
{
   if(!magical) return;
    spell_points[0] += h;

    return spell_points[0];
}

query_ghost() { return ghost; }

query_playing_time() 
{ 
  return playing_time;
}

show_playing_time() {
write("Playing time: "+"closed/mcp"->stime(time()-playing_time)+".\n");
}

/* This is used by the shop etc. */

add_credits(m) 
{
    if(credits+m<0.0) return 0;

    credits+=m+0.0;

    if(wiz_level) return 1;

    if(test_player(this_object())) add_worth(m);

#ifdef LOG_MONEY

    if(this_player() && this_player()!=this_object() &&
       this_player()->query_wiz_level())
             log_file("MONEY",this_player()->query_real_name()+" gave "+name+
                      " "+m+" coins in room: "+file_name(environment())+
                      " on: "+ctime(time())+"\n");

#endif

  return 1;
}

query_credits() { return credits; }

query_exp(cls) {
int n;

  if(!cls) return exp;
    
  n=search_array(classes,cls);
  return exp[n];

}

static int busy_run_away;

run_away() 
{
    object here;
    int i, j;
    string dest_dir;
    
    if(busy_run_away) return;
    busy_run_away=1;
    here = environment();
	tell_object(this_object(),
	  "You receive a devastating blow from "+this_player()->query_name()+"!\n"+
	  "Your legs run away with you!\n");
	if(whimpy_dir) {
		tell_object(this_object(),
			"You attempt the emergency command '"+whimpy_dir+"'.\n");
		command(whimpy_dir);
		whimpy_dir = 0;
                if(here!= environment())
				 {
                    tell_object(this_object(),"And you make it!\n");
					busy_run_away=0;
					return;
				 }
	}
    if(here==environment()) dest_dir=environment()->query_dest_dir();
    if(dest_dir) command(dest_dir[(random(sizeof(dest_dir)/2)*2)+1]);
    for(i=0;i<17 && here == environment();i++) {
        j=random(17);
	command(DIR_ARRAY[j]);

    }

    if (here == environment()) {
	say(cap_name + " tried, but failed to run away.\n", this_object());
	tell_object(this_object(), "But failed!\n");
    } else {
	tell_object(this_object(), "And you made it!\n");
    }
  busy_run_away=0;
}

query_hit_points(n) 
{
    if(n==-1) return hit_points;

    return hit_points[n];
}

query_wimpy() {
    return whimpy;
}

query_spell_points(n) 
{
    if(!magical) return;
   
    if(n==-1) return spell_points;

    return spell_points[n];
}

query_weilded_weapons(n)
{
    if(!pointerp(weapons) || n> sizeof(weapons)-1 ) return 0;

    if(n<0) return weapons;
    
    return weapons[n];
}

query_ac(n) 
{
    if(n<0) return armour_classes;

    return armour_classes[n];
}

query_age() { return age; }

query_gender() { return gender; }
query_neuter() { return !gender; }
query_male() { return gender == 1; }
query_female() { return gender == 2; }

set_gender(g)
{
    if(g>=0 && g<=2) gender = g;
}

set_neuter() { gender = 0; }
set_male() { gender = 1; }
set_female() { gender = 2; }

query_gender_string(flag)
{
   switch(flag)
     {
       default:
	    if (!gender) return "neuter";
	      else if (gender == 1) return "male";
    		else return "female";
	case 1:
	    if (!gender) return "it";
	      else if (gender == 1) return "he";
	        else return "she";
	case 2:
    	    if (!gender) return "its";
	      else if (gender == 1) return "his";
	       else return "her";
	case 3:
    	    if (!gender) return "it";
	      else if (gender == 1) return "him";
  	        else return "her";
      }
}

query_stats() { return ({Str[0],Dex[0],Int[0],Wis[0],Con[0]}); }

query_hit_points_die(n) 
{
  if(n==-1) return hit_points_die;

  return hit_points_die[n];
}

query_bonus() { return Str[0]/6;} 
query_attacker() { return attacker;}
set_attacker(at) { attacker=at;}
query_num_attacks() { return num_attacks;}
set_num_attacks(n) { if(n>0) num_attacks=n; }
query_race() { return race; }
query_race_min_max() { return race_min_max; }

set_race(rc,mon_rc)
{
int i,c;

 if(!stringp(rc) && !stringp(mon_rc)) return 0;

 if(rc)    /* for player and npc types */
  {
     rc=lower_case(rc);

     if((i=search_array(RACE_ARRAY,rc))==-1) return 0;

     race_min_max=MCP->strip_lf(read_file("/r1/obj/data/race_min_max",i+1,1));

     race_min_max=explode(race_min_max,',');

     for(i=sizeof(race_min_max);c<i;c++)
      {
         race_min_max[c]=explode(race_min_max[c],':');
         race_min_max[c][0]=atoi(race_min_max[c][0]);
         race_min_max[c][1]=atoi(race_min_max[c][1]);
      }

      return race=rc;
   }

  /*  for monster types */
  return race=mon_rc;
 
}

add_class(cls)
{
  string tmp;
  int ln,i;

  if(classes[2] || (ln=search_array(CLASS_ARRAY,cls))==-1 ||
     query_class(cls) || query_class("monster") || 
     (cls=="monster" && classes[0]) ) return 0;

  tmp=MCP->strip_lf(read_file("/r1/obj/data/classes",++ln,1));
  tmp=explode(tmp,',');tmp[1]=atoi(tmp[1]);tmp[2]=atoi(tmp[2]);

  if(hit_points_die[0]<tmp[1]) hit_points_die[0]=tmp[1];
  if(hit_points_die[1]<tmp[2]) hit_points_die[1]=tmp[2];

  if(!classes[0]) classes[0]=cls;
	else if(!classes[1]) classes[1]=cls;
	  else classes[2]=cls;

  if(cls=="monster") return 1;

   /* for players and npc's only. */

  ln=search_array(CLASS_ARRAY,cls)+1;

  tmp=MCP->strip_lf(read_file("/r1/obj/data/class_min_max",ln,1));
  tmp=explode(tmp,',');
  
  if(!class_min_max)
    {
       class_min_max=tmp;

       for(;i<5;i++)
	{
	  class_min_max[i]=explode(class_min_max[i],':');
	  class_min_max[i][0]=atoi(class_min_max[i][0]);
	  class_min_max[i][1]=atoi(class_min_max[i][1]);
        }
    }
    else 
      {
	  for(;i<5;i++)
           {
	     tmp[i]=explode(tmp[i],':');
	     tmp[i][0]=atoi(tmp[i][0]);
	     tmp[i][1]=atoi(tmp[i][1]);

             if(class_min_max[i][0]<tmp[i][0])
			class_min_max[i][0]=tmp[i][0];

	     if(class_min_max[i][1]<tmp[i][1])
			class_min_max[i][1]=tmp[i][1];
	   }
      }

  return 1;
}

query_class_min_max() { return class_min_max; }

query_class(sn)
{
  if(stringp(sn))
    if(search_array(classes,sn)!=-1) return 1;
	else return 0;

  if(!numberp(sn)) return -1;

  if(!sn) return classes;

  return classes[sn];
}

hit(n,ob)
{
  hit_points[0]-=n;

  if(hit_points[0]<=-10)
    {
      if(ob) tell_object(ob,"You have defeated "+name+".\n");
      death();
      return;
    }

  if(hit_points[0]<=0)
   {
/* 
      if(!check_system_shoock())
	{
	  tell_object(this_object(),"You fall unconseous.\n");
	  tell_object(ob,"You have renderd your apponent unconseous.\n");
          temp_remove_attacks();
	  block_actions();
	  knocked_out=1;
        }
*/
   }
}

query_dex(n)
{
  if(n==-1) return Dex; 

   return Dex[n];
}

set_dex(n)
{
   string tmp;
   int i,s;

   if(!numberp(n) || !race_check(n,1) || !class_check(n,1) || n>25) 
		return 0;

   n=get_bonus_penalty(n,1);

   tmp=MCP->strip_lf(read_file("/r1/obj/data/dextable.1",n,1));

   if(Dex[3])
     {
       armour_classes[0]+=Dex[3];
       armour_classes[1]+=Dex[3];
     }

   Dex=({n})+explode(tmp,',');

   for(s=sizeof(Dex),i=1;i<s;i++)
      Dex[i]=atoi(Dex[i]);

   if(Dex[3])
     {
       armour_classes[0]-=Dex[3];
       armour_classes[1]-=Dex[3];
     }

   return n;
}

query_str(n)
{
  if(n==-1) return Str;

  return Dex[n];
}

set_str(n)
{
   string tmp;
   int s,i;

   if(!numberp(n) || !race_check(n,2) || !class_check(n,2) || n>25) 
	return 0;

   n=get_bonus_penalty(n,2);

   if(n<19 || (n>=19 && !query_class("fighter")))
	tmp=MCP->strip_lf(read_file("/r1/obj/data/strtable.1",n,1));
     else tmp=MCP->strip_lf(read_file("/r1/obj/data/strtable.2",n-18,1));

   Str=({n})+explode(tmp,',');

   for(i=1,s=sizeof(Str);i<s;i++)
     Str[i]=atoi(Str[i]);

   return n;
}

query_int(n)
{ 
  if(n==-1) return Int;

  return Int[n];
}

set_int(n)
{
   string tmp;
   int i,s;

   if(!numberp(n) || !race_check(n,3) || !class_check(n,3) || n>25)
	 return 0;

   if(n<9) tmp="0,0,0,0,0";
     else tmp=MCP->strip_lf(read_file("/r1/obj/data/inttable.1",n-8,1));

   Int=({n})+explode(tmp,',');

   for(i=1,s=sizeof(Int);i<s;i++)
      Int[i]=atoi(Int[i]);

   return n;
}

query_wis(n)
{
  if(n==-1) return Wis;

  return Wis[n];
}

set_wis(n)
{
   string tmp;
   int i,s;

   if(!numberp(n) || !race_check(n,3) || !class_check(n,3) || n>25)
	 return 0;

   tmp=MCP->strip_lf(read_file("/r1/obj/data/wistable.1",n,1));

   Wis=({n})+explode(tmp,',');

   for(i=1,s=sizeof(Wis);i<s;i++)
     Wis[i]=atoi(Wis[i]);

   return n;
}

query_con(n)
{
   if(n==-1) return Con;

   return Con[n];
}

set_con(n)
{
   string tmp;
   int i,s;

   if(!numberp(n) || !race_check(n,4) || !class_check(n,4) || n>25) 
	return 0;

   n=get_bonus_penalty(n,4);

   if(n<17 || (n>=17 && !query_class("fighter")))
	tmp=MCP->strip_lf(read_file("/r1/obj/data/contable.1",n,1));
     else tmp=MCP->strip_lf(read_file("/r1/obj/data/contable.2",n-16,1));

   Con=({n})+explode(tmp,',');

   for(i=1,s=sizeof(Con);i<s;i++)
     Con[i]=atoi(Con[i]);

   return n;
}

get_bonus_penalty(num,stat)
{
 int i,len;
 string tmp;

 tmp=MCP->stirp_lf(read_file("/r1/obj/data/races",i,8));

 for(len=strlen(race);tmp;tmp=strchr(tmp,'\n'))
   {
     tmp++;
     if(tmp[0]=0) break; 
     if(!strncmp(tmp,race,len))
       { 
	  string s;

	  s=strchr(tmp,'\n');
	  tmp[strlen(tmp)-strlen(s)+1]=0;
	  tmp=explode(tmp,',');

 	  switch(stat)
	   {
	       case 1:
               case 2:
			num+=tmp[stat];
			break;
 	       case 4:
			num+=tmp[3];
			break;
	      default:
	   }
	   break;
       }
    }
  
  return num;
}

race_check(n,stat)
{
   if(this_object()->query_monster()) return 1;

   if(n<race_min_max[stat-1][0] || n>race_min_max[stat-1][1]) return 0;

   return 1;
}

class_check(n,stat)
{
   if(this_object()->query_monster()) return 1;

   if(n<class_min_max[stat-1][0] || n>class_min_max[stat-1][1]) return 0;

   return 1;
}

show_stats()
{
int i;
string tmp;

  write(short() + "\n");
  write("Alignment   : "+alignment+"\n");
  write("Gender      : "+query_gender_string()+"\n");
  write("Race        : "+race+"\n");
  write("Class       : "+classes[0]);
    if(classes[1]) write("/"+classes[1]);
    if(classes[2]) write("/"+classes[2]);
  write("\n");
  write("Level       : " + levels[0]);
    if(levels[1]) write("/"+levels[1]);
    if(levels[2]) write("/"+levels[2]);
    write("\n");
   write(sprintf(tmp,"Strength    : %-2d     Dexterity   : %-2d     Constitution: %-2d\n",Str[0],Dex[0],Con[0]));
   write(sprintf(tmp,"Intelligence: %-2d     Wisdom      : %-2d\n",Int[0],Wis[0]));
   write("Credits     : "+credits+"\n");
   if(magical)
      write(sprintf(tmp,"Spell Points: %-3d     Max        : %-3d\n",spell_points[1],spell_points[0]));
      write(sprintf(tmp,"Hit Points  : %-3d     Max        : %-3d\n",hit_points[1],hit_points[0]));
   if(test_player(this_object()))
     {
        write("Experience  : "+exp[0]);
        if(exp[1]) write("/"+exp[1]);
        if(exp[2]) write("/"+exp[2]);
        write("\n");
      }
     else write("Experience  : "+query_exp_value()+"\n");
   write("# of Attacks: "+num_attacks+"\n");
   write("Armor Class : "+armour_classes[0]+"/"+armour_classes[1]+"/"+armour_classes[2]+"\n");
  if(armour)
  {
   write("Armor       : ");
	for(i=0;i<sizeof(armour);i++)
	  {
		write(armour[i]->short());
		if(i<sizeof(armour)-1) write(", ");
	  }
    write("\n");
  }
  if(weapons)
   {
     write("Weapon(s)   : "+weapons[0]->short());
	 if(sizeof(weapons)==2) write(", "+weapons[1]->short());
	 write("\n");
   }
   write(sprintf(tmp,"Weight      : %-9.4f    Carrying   : %-9.4f   Encumbrance : %8.4f%%\n",local_weight[0],local_weight[1],(query_encumbrance()*100)));
   write("Height      : "+height+"\n");
   write("Age         : "+age+"\n");
   write("Skills      : "+skill_names+"\n");
   write("              "+skill_levels+"\n");

   if(!quests) write("Quests      : "+quests+"\n");
   show_playing_time();
   if(creator(this_object()))
	  write("Creator      : "+creator(this_object())+"\n");
}

set_level(cls,l)
{
int i;

 if(!stringp(cls) || (i=search_array(classes,cls))==-1) return;

 if(l<0) l=1;
 
 return (levels[i]=l);
}

add_exp(cls,num)
{
 int n;

 n=search_array(classes,cls);
 exp[n]+=num;
}

set_hit_points(hp)
{
   hit_points=({hp,hp});
}

set_spell_points(sp) 
{
  if(!magical) return;
  spell_points=({sp,sp});
}

query_exp_value()
{
  return MCP->exp_value(this_object());
}

set_age(n)
{
 if(n<1) return;

 age=n+0.0;
}

set_body_weight(n)
{
 if(n<1) return;

 local_weight[0]=n+0.0;
}

query_weight(n)
{
  if(n<0) return local_weight;

  if(!n) return local_weight[0]+local_weight[1];

  return local_weight[n-1];
}

query_encumbrance()
{
  return local_weight[1]/Str[3];
}

set_height(n)
{
  if(n<1) return;
  height=n+0.0;
}

query_height() { return height; }

set_can_talk(n) { can_talk=n; }

query_can_talk() { return can_talk; }

query_num_hands() { return num_hands;}

set_num_hands(h) { num_hands=h; }

set_name(nm) 
{
  if(!MCP->valid_name(nm)) return 0;

  cap_name=capitalize((name=lower_case(nm)));
}

set_magical(n) { magical=n; }

query_magical() { return magical;}

query_invis() { return is_invis; }

set_invis(n) { is_invis=n; }

query_use_light() { return use_light; }

query_hunger() { return hunger; }
query_drunk()  { return drunk;  }
query_thirst() { return thirst; }

add_food(n)
{
    if(hunger>=MAX_FOOD) return;
    return hunger+=n;
}

add_fluid(n)
{
   if(thirst>=MAX_FLUID) return;
   return thirst+=n;
}

add_alcohol(n)
{
   if(drunk>=MAX_DRUNK) return;
   return drunk+=n;
}

/* skill names should be limited to 15 characters. */

add_skill(new_skill)
{
  if(!pointerp(skill_levels)) { skill_names=({});skill_levels=({}); }

  if(search_array(skill_names,new_skill[0])!=-1) return 0;

  if(new_skill[1]>25) new_skill[1]=25;
   else if(new_skill[1]<1) return 0;

  skill_names+=({new_skill[0]});
  skill_levels+=({new_skill[1]});

  return 1;
}

remove_skill(old_skill)
{
int i;

  if(!pointerp(skill_names))  return 0;

  if((i=search_array(skill_names,old_skill))==-1) return 0;

  skill_names=del_element(skill_names,i);
  skill_levels=del_element(skill_levels,i);

  return 1;
}

increase_skill(i_skill)
{
  int i;

  if(!pointerp(skill_levels)) return 0;

  if((i=search_array(skill_names,i_skill[0]))!=-1)
  {
    if(i_skill[1]<1) remove_skill(i_skill[0]);
      {
         skill_levels[i]=i_skill[1];
         if(skill_levels[i]>25) skill_levels[i]=25;
      }
    return 1;
  }
}

query_skill(q_skill)
{
  int i;

  if(!pointerp(skill_levels)) return 0;

  if((i=search_array(skill_names,q_skill))!=-1)
     return ({q_skill,skill_levels[i]});
}
