/*
 * Definition of an object.
 *
 * The reset is used as follows:
 * 0: There is an error in the reset() in this object. Never call it again.
 * 1: Normal state.
 * 2 or higher: This is an interactive player, that has not given any commands
 *		for a number of reset periods.
 *
 * In the beginning, 'prog' points to the program tree, and 'status' to the
 * linked list of variable definitions.
 * After the object has been loaded back from the swap, the program tree
 * and variable definitions will be stored in one sequential memory area.
 * This area will then be pointed to by 'block'.
 */

#define VALID_OB        0x0001  /* This flag is removed if destruct() is
                                   called on this object.                   */
#define UNUSED		0x0002	/* If we have a reset in this object        */
#define COMMANDS	0x0004	/* If this object responds to commands      */
#define CLONE		0x0008	/* If this object is a clone		    */
#define PRIVATE		0x0010	/* If this is a non-inheritable object      */
#define GLOBAL		0x0020	/* If this is non-realm object		    */
#define VERB_MATCHING	0x0040	/* If this uses patern matchs on verbs      */
#define BLOCK_SHOUTS 	0x0080	/* If this object does not hear shouts      */
#define SWAPPED		0x0100	/* If this object is swapped		    */
#define NO_CLONE	0x0200	/* We can not be cloned                     */
#define RESET		0x0400  /* We MUST be reset each time.              */
#define NO_EXT_SAVE     0x0800  /* We can not be saved externally	    */
#define MOVE		0x1000	/* We are in the first-half of move_object  */
#define MOVE2		0x2000  /* We are in the second-half of move_object */
#define DEST		0x4000  /* Destructed				    */
#define PLAYER		0x8000  /* Is this a player object		    */

#define NUM_REF_OB	6

struct link_list {
	struct fun_p *fun_p;		/* for heart_beat use */
	struct object *ob;
	struct link_list *next;
	struct link_list *prev;
};

struct object {
	unsigned short flags;		/* See list of flag defines above.    */
	unsigned char not_touched;	/* # of resets the object has not been
					   touched.			      */
	short total_light;		/* The total light affecting this
					   object.			      */
	unsigned short num_variables;	/* Number of variables 		      */
	int swap_num;			/* Swap file offset. -1 is not swapped
					   yet.				      */

	unsigned int num_ref;	/* Number of references to this object.  We can
				   not free the memory this object uses until
				   all the refernces are gone.   But we can
				   free up all variables and pointers when the 
				   object is destroid.                        */
	unsigned int swap_size;	/* Size of object on swap area. 	      */
	char *logged;		/* File to log input/output of this object to */
	char *block;		/* Pointer to the start of the prog block     */
	char *name;		/* Name of object 			      */
	char *living_name;	/* Name of living object if in hash	      */
	struct realms *realm;	/* Realm mark for all objects.		      */
	struct lnode_def *prog;	/* The program itself. 			      */

	struct lnode_var_def *status;	  /* The variable list		      */
	struct interactive *interactive;  /* Data about an interactive player */
	struct sentence **verbs;  	  /* Action verbs for this object     */
	struct ed_buffer *ed_buffer;	  /* Local ed			      */
	struct wiz_list *wl;		  /* What wizard defined this object  */
	struct inherit_list *inherit;	  /* Where to inherit from	      */
	struct link_list *links[2];	  /* heart beats and resets           */
	struct fun_p *reset;		  /* so this can be called at any time*/
	struct fun_p *fun_p[10];          /* function pointers for speed.     */
	struct object *ref_ob[NUM_REF_OB];/* Object pointers

					     0 - Object containing us.
					     1 - The first object in us.
					     2 - If we are in an inventory,
						 the object after us.
					     3 - If we are in an inventory,
						 the object before us.
					     4 - next object in destruct or
						  free list.
					     5 - move_object destination.
									     */
	struct svalue *variables;

		/* The variables must come last in the struct */
};


struct MCP {
	OBJECT *ob; 
	struct fun_p *fun_p[5];
};

extern struct object *load_object PROT((struct realms *,char *)), *find_object PROT((char *));
extern struct object *get_empty_object(), *find_object PROT((char *)),
		     *find_object2 PROT((char *,int));
extern struct object *current_object, *command_giver;

extern struct object *obj_free_list;

void destruct_object PROT((struct object *,char *));
void remove_destructed_objects(), 
     move_object PROT((struct object *, struct object *,int,int)),
     tell_object PROT((struct object *, VALUE *)),
     tell_npc PROT((struct object *, char *)),
     add_ref PROT((struct object *, char *));

struct value *clone_object PROT((char *));

int restore_object PROT((struct object *, char *,int)),
    save_object PROT((struct object *, char *,int)),
    save_inventory PROT((struct object *,char *)),
    check_object PROT((struct object *)),
    restore_inventory PROT((struct object *,char *));

extern struct MCP mcp;
