#include <stdio.h>
#include <string.h>
#include <errno.h>
#include "lint.h"
#include "wiz_list.h"
#include "lnode.h"
#include "object.h"
#include "lint2.h"
#include "realms.h"
#include "lang.h"
#include "error.h"

/*
 * Maintain the wizards high score list about most popular castle.
 */

extern char *string_copy PROT((char *)), *xalloc PROT((int));

extern void add_message();

struct wiz_list *all_wiz;

/*
 * Sort the wiz list in ascending order.
 */

static struct wiz_list *insert(w, wl)
    struct wiz_list *w, *wl;
{
    if (!wl) {
	w->next = 0;
	return w;
    }
    if (w->score > wl->score) {
	wl->next = insert(w, wl->next);
	return wl;
    }
    w->next = wl;
    return w;
}

static void rebuild_list() {
    struct wiz_list *wl, *w, *new_list = 0;

    for (w = all_wiz; w; w = wl) {
	wl = w->next;
	new_list = insert(w, new_list);
    }
    all_wiz = new_list;
}

/*
 * Find the data, if it exists.
 */
struct wiz_list *find_wiz(name,realm)
    char *name,*realm;
{
    int length;
    struct wiz_list *wl;

    if(!name || !realm) return 0;
    for (wl = all_wiz; wl; wl = wl->next)
        if (!strcmp(wl->name, name))
	  {
	    REALMS *f_realm;
	    extern REALMS *sys_realm;
	    f_realm=find_realm(realm,1);
	    if(!f_realm) return 0;
	    if(!wl->realm)
          if(f_realm==sys_realm)
			{
				wl->realm=f_realm;
				return wl;
			}
	      else return 0;
	    if(wl->realm==f_realm) return wl;
	  }
    return 0;
}

/*
 * Check that a name exists. Add it, if it doesn't.
 */
struct wiz_list *add_name(str,realm)
    char *str,*realm;
{
    struct wiz_list *wl;

    wl = find_wiz(str,realm);
    if (wl) return wl;
    wl = (struct wiz_list *)xalloc(sizeof (struct wiz_list));
    str = string_copy(str);
    wl->name = str;
    wl->realm=find_realm(realm,1);
    wl->length = strlen(str);
    wl->score = 0;
    wl->cost = 0;
    wl->heart_beats = 0;
    wl->total_worth = 0;
    wl->next = all_wiz;
    wl->size_array = 0;
    wl->file_name = 0;
    wl->error_message = 0;
    all_wiz = wl;
    return wl;
}

char *extract_wizard(path)
   char *path;
{
  char *w,*p, *wiz;
  int wl,rl;
  REALMS *realm;
  extern REALMS *sys_realm;

  if(path[0]=='/') path++;

  realm=extract_realm(path);

  if(realm==sys_realm) return 0;

  rl=strlen(realm->realm_dir)+1;

  path+=rl;

  if(strncmp(path,"w/",2)) return 0;

  path+=2;

  w=string_copy(path);

  p=strchr(w,'/');

  if(p)
   {
     p[0]=0;
     wl=strlen(w);
     p[0]=' ';
   }
  else wl=strlen(w);

  wiz=xalloc(wl+1);

  strncpy(wiz,w,wl);

  wiz[wl]=0;

  free(w);

  return wiz;
}

/*
 * Add score to an existing name.
 */
void add_score(name,realm, score,total_worth)
    char *name,*realm;
    int score,total_worth;
{
    struct wiz_list *wl;

    wl = find_wiz(name,realm);
    if (!wl)
       debug_message(sys_realm,"Can not find wizard in wizlist: %s\n",name);
     else
      {
        wl->score += score;
        wl->total_worth+=total_worth;
      }
}

/*
 * This one is called at every complete walkaround of reset.
 */
void wiz_decay() {
    struct wiz_list *wl;

    for (wl = all_wiz; wl; wl = wl->next) {
        wl->score = wl->score * 99 / 100;
	wl->total_worth = wl->total_worth * 99 / 100;
	wl->cost = wl->cost * 9 / 10;
	wl->heart_beats = wl->heart_beats * 9 / 10;
    }
    save_wiz_file();
}

/*
 * Load the wizlist file.
 */
void load_wiz_file()
{
    char buff[79];
    FILE *f;

    extern int atoi();

    f = fopen("info/Wizard_list", "r");
    if(!f && errno == EMFILE) fatal("Max files open.\n");
    if (f == NULL)
        return;
    while(fgets(buff, sizeof buff, f) != NULL) {
        char p[12],name[15],realm[40], q[12];

        sscanf(buff,"%s %s %s %s\n",realm,name,p,q);
	if (p == '\0' && q=='\0') {
	    debug_message(sys_realm,"Bad Wizard_list file.\n");
	    break;
	}
	if (atoi(p) != 0 || atoi(q) !=0) {
	    (void)add_name(name,realm);
	    add_score(name,realm,atoi(p),atoi(q));
	}
    }
    fclose(f);
}

/*
 * Save the wizlist file.
 */
void save_wiz_file()
{
    struct wiz_list *wl;
    FILE *f;

    if(!all_wiz) return;
    f = fopen("info/Wizard_list", "w");
    if(!f && errno == EMFILE) fatal("Max files open.\n");
    if (f == NULL) {
        debug_message(sys_realm,"Could not open Wizard_list for write\n");
        return;
    }
    for (wl = all_wiz; wl; wl = wl->next)
        if(wl->score >0 || wl->total_worth>0)
	      if(wl->realm)
            fprintf(f, "%s %s %d %d\n",wl->realm->realm_owner, wl->name, wl->score, wl->total_worth);
		  else debug_message(sys_realm,"wl without a realm: %s\n",wl->name);
    fclose(f);
}

void wizlist(name,realm)
    char *name,*realm;
{
    struct wiz_list *wl, *this_wiz;
    int total = 0, tot_worth=0, num = 0, quot=0, this_pos, total_wizards;
    int tot_quot=0, tot_worth2=0,total2=0, quot2=0;
    extern struct object *command_giver;
    struct realms *rn;
    int all = 0;

    rn=find_realm(realm,1);

    if (!command_giver || !all_wiz) return;
    if (!name)
      {
		VALUE *v;

		v = apply(0,"query_user_name", command_giver, &const0,0);
        if (strchk(v,1)) return;
		name=v->u.string;
      }
    if (!strcmp(name, "all")) all = 1;
    this_wiz = find_wiz(name,realm);
    rebuild_list();
    for (num = 0, wl = all_wiz; wl; wl = wl->next) 
	  {
		if(wl->realm!=rn) continue;
        total += wl->score;
		if(strcmp(wl->name,"mud"))
          {
	     	total2+=wl->score;
	     	tot_worth2+=wl->total_worth;
	  	  }
        num++;
        tot_worth+=wl->total_worth;
		if (wl == this_wiz) this_pos = num;
      }
    total_wizards = num?num-1:0;
    add_message("\nWizard list for realm: %s\n\n",rn->realm_name);
    add_message("%-15s %-6s  %-7s %-5s    %%     %-6s  %-6s  %s\n\n","Wizard",
			    "Points","Exp+Mo","Ratio","Nodes","HeartB","Array Aloc");
    this_pos = num - this_pos + 1;
    for (wl = all_wiz; wl; wl = wl->next) 
	   {
		 if(wl->realm!=rn) continue;
		 if (!all && num > 13 && (num < this_pos - 2 || num > this_pos + 2)) ;
	   	  else
           {
            if(wl->total_worth>0 && wl->score>0) quot=wl->total_worth/wl->score;
			   else quot=0;
            add_message("%-15s %6d %8d %5d %6.2f%% [%6dk,%6d] %4d\n",wl->name,
						wl->score,wl->total_worth,quot,
						total?(float)(wl->score*100)/(float)total:(float)0,
                        wl->cost/1024,wl->heart_beats,wl->size_array);
           }
		 num--;
       }
    if(tot_worth2>0 && total2>0) quot2=tot_worth2/total2;
    add_message("\nTotals:\nWizards (%3d):  %6d %8d %5d %6.2f%%\n",
			    total_wizards,total2,tot_worth2,quot2,
				total?(float)(total2*100)/(float)total:(float)0);
    if(tot_worth>0 && total>0) tot_quot=tot_worth/total;
    add_message("Overall      :  %6d %8d %5d 100.00%%\n\n",total, tot_worth,tot_quot);
}

void remove_wiz_list() {
    struct wiz_list *wl, *w;

    for (w = all_wiz; w; w = wl) {
	free(w->name);
	wl = w->next;
	free((char *)w);
    }
}

void save_error(msg, file, line,realm_dir)
    char *msg, *realm_dir;
    char *file;
    int line;
{
    struct wiz_list *wl;
    REALMS *r;
    char name[100];
    char *p;
    int len;

    if (sscanf(file, "w/%s/", name) != 1) return;
    
    p = strchr(name, '/');
    if (p) *p = '\0';
    r=find_realm(realm_dir,2);
    if(!r) return;
    wl = add_name(name,r->realm_owner);
    if (wl->file_name)
	free(wl->file_name);
    len = strlen(file);
    wl->file_name = xalloc(len + 3); /* May add .c plus the null byte */
    strcpy(wl->file_name, file);
    /*
     * If it is a cloned object, we have to find out what the file
     * name is, and add '.c'.
     */
    p = strrchr(wl->file_name, '#');
    if (p) {
	p[0] = '.';
	p[1] = 'c';
	p[2] = '\0';
	len = p - wl->file_name + 2;
    }
    if (wl->file_name[len-1] != 'c' || wl->file_name[len-2] != '.')
	strcat(wl->file_name, ".c");
    if (wl->error_message)
	free(wl->error_message);
    wl->error_message = string_copy(msg);
    wl->line_number = line;
}

char *get_error_file(name,realm)
    char *name, *realm;
{
    struct value *v;
    extern struct object *command_giver;
    struct wiz_list *wl;

    wl = add_name(name,realm);
    /*
     * The error_message is used as a flag if there has been any error.
     */
    if (wl->error_message == 0) {
	add_message("No error.\n");
	return 0;
    }
    add_message("%s:(%d) : %s\n", wl->file_name, wl->line_number,
		wl->error_message);
    free(wl->error_message);
    wl->error_message = 0;
    return wl->file_name;
}
